#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################

from typing import Optional, Tuple

try:
    import open3d as o3d
except ModuleNotFoundError:
    pass
from tardis_em.utils import SCANNET_COLOR_MAP_20


import matplotlib.pyplot as plt
import numpy as np


def img_is_color(img):
    if len(img.shape) == 3:
        # Check the color channels to see if they're all the same.
        c1, c2, c3 = img[:, :, 0], img[:, :, 1], img[:, :, 2]
        if np.all(c1 == c2) and np.all(c2 == c3):
            return True

    return False


def show_image_list(
    list_images,
    list_titles=None,
    list_cmaps=None,
    list_mask_cmaps=None,
    grid=True,
    num_cols=2,
    figsize=(20, 10),
    title_fontsize=30,
    list_masks=None,
    dpi=100,
):
    """
    Shows a grid of images, where each image is a Numpy array. The images can be either
    RGB or grayscale.

    Args:
        list_images (list): List of the images to be displayed.
        list_titles (list or None): Optional list of titles to be shown for each image.
        list_cmaps (list, str or None): Optional list of cmap values for each image.
            If None, then cmap will be automatically inferred.
        list_mask_cmaps (list, str or None):
        grid (boolean): If True, show a grid over each image
        num_cols (int): Number of columns to show.
        figsize (tuple): Value to be passed to pyplot.figure()
        title_fontsize (int): Value to be passed to set_title()
        list_masks(list, None):
        dpi (int):
    """
    assert isinstance(list_images, list)
    assert len(list_images) > 0
    assert isinstance(list_images[0], np.ndarray)

    if list_titles is not None:
        assert isinstance(list_titles, list)
        assert len(list_images) == len(list_titles), "%d imgs != %d titles" % (
            len(list_images),
            len(list_titles),
        )

    if list_cmaps is not None:
        assert isinstance(list_cmaps, (list, str))
        if not isinstance(list_cmaps, str):
            assert len(list_images) == len(list_cmaps), "%d imgs != %d cmaps" % (
                len(list_images),
                len(list_cmaps),
            )

    if list_masks is not None:
        assert len(list_masks) == len(list_images)

    num_images = len(list_images)
    num_cols = min(num_images, num_cols)
    num_rows = int(num_images / num_cols) + (1 if num_images % num_cols != 0 else 0)

    # Create a grid of subplots.
    fig, axes = plt.subplots(num_rows, num_cols, figsize=figsize, dpi=dpi)

    # Create list of axes for easy iteration.
    if isinstance(axes, np.ndarray):
        list_axes = list(axes.flat)
    else:
        list_axes = [axes]

    for i in range(num_images):
        img = list_images[i]
        if list_masks is not None:
            mask = list_masks[i]

        if isinstance(list_cmaps, str):
            cmap = list_cmaps
        else:
            cmap = (
                list_cmaps[i]
                if list_cmaps is not None
                else (None if img_is_color(img) else "gray")
            )

        if list_mask_cmaps is not None:
            if isinstance(list_mask_cmaps, str):
                cmap_mask = list_mask_cmaps
            else:
                cmap_mask = (
                    list_mask_cmaps[i] if list_mask_cmaps is not None else "Reds"
                )
        else:
            cmap_mask = "Reds"

        list_axes[i].imshow(img, cmap=cmap)
        if list_masks is not None:
            list_axes[i].imshow(mask, cmap=cmap_mask, alpha=0.5)

        if title_fontsize is not None:
            list_axes[i].set_title(
                list_titles[i] if list_titles is not None else "Image %d" % i,
                fontsize=title_fontsize,
            )
        list_axes[i].grid(grid)

    for i in range(num_images, len(list_axes)):
        list_axes[i].set_visible(False)

    fig.tight_layout()
    plt.show()


def _dataset_format(coord: np.ndarray, segmented: bool) -> Tuple[np.ndarray, bool]:
    """
    Silently check for an array format and correct 2D datasets to 3D.

    Args:
        coord (np.ndarray): 2D or 3D array of shape [(s) x X x Y x Z] or [(s) x X x Y].
        segmented (bool): If True expect (s) in a data format as segmented values.

    Returns:
        Tuple[np.ndarray, bool]: Checked and corrected coord array with boolean
        statement if array is compatible.
    """
    check = True

    if segmented:
        if coord.shape[1] not in [3, 4]:
            check = False
            print("Coord data must be 2D/3D with labels (4D/5D)")

        # Correct 2D to 3D
        if coord.shape[1] == 3:
            coord = np.vstack(
                (coord[:, 0], coord[:, 1], coord[:, 2], np.zeros((coord.shape[0],)))
            ).T
    else:
        if coord.shape[1] not in [2, 3]:
            check = False
            print("Coord data must be 2D/3D with labels (2D/3D)")

        # Correct 2D to 3D
        if coord.shape[1] == 2:
            coord = np.vstack((coord[:, 0], coord[:, 1], np.zeros((coord.shape[0],)))).T

    return coord, check


def _rgb(coord: np.ndarray, segmented: bool, ScanNet=False, color=False) -> np.ndarray:
    """
    Convert float to RGB classes.

    Use predefined Scannet V2 RBG classes or random RGB classes.

    Args:
        coord (np.ndarray): 2D or 3D array of shape [(s) x X x Y x Z] or [(s) x X x Y].
        segmented (bool): If True expect (s) in a data format as segmented values.
        ScanNet (bool): If True output scannet v2 classes.

    Returns:
        np.ndarray: 3D array with RGB values for each point.
    """
    rgb = np.zeros((coord.shape[0], 3), dtype=np.float64)

    if segmented:
        if ScanNet:
            for id_, i in enumerate(coord[:, 0]):
                color = SCANNET_COLOR_MAP_20.get(i, SCANNET_COLOR_MAP_20[0])
                rgb[id_, :] = [x / 255 for x in color]
        else:
            unique_ids = np.unique(coord[:, 0])
            if color:
                rgb_list = [np.array((0, 0, 0)) for _ in unique_ids]
            else:
                rgb_list = [
                    np.array((np.random.rand(), np.random.rand(), np.random.rand()))
                    for _ in unique_ids
                ]
            id_to_rgb = {idx: color for idx, color in zip(unique_ids, rgb_list)}

            for id_, i in enumerate(coord[:, 0]):
                df = id_to_rgb[i]
                rgb[id_, :] = df
    else:
        rgb[:] = [1, 0, 0]

    return rgb


def segment_to_graph(coord: np.ndarray) -> list:
    """
    Build filament vector lines for open3D.

    Args:
        coord (np.ndarray): 2D or 3D array of shape [(s) x X x Y x Z] or [(s) x X x Y].

    Returns:
        list: list of segments converted for open3D
    """
    graph_list = []
    stop = 0

    for i in np.unique(coord[:, 0]):
        id_ = np.where(coord[:, 0] == i)[0]
        id_ = coord[id_]

        x = 0  # Iterator checking if current point is a first on in the list
        start = stop
        stop += len(id_)

        if x == 0:
            graph_list.append([start, start + 1])

        length = stop - start  # Number of point in a segment
        for j in range(1, length - 1):
            graph_list.append([start + (x + 1), start + x])

            if j != (stop - 1):
                graph_list.append([start + (x + 1), start + (x + 2)])
            x += 1
        graph_list.append([start + (x + 1), start + x])

    return graph_list


def VisualizePointCloud(
    coord: np.ndarray, segmented: bool, rgb: Optional[np.ndarray] = None, animate=False
):
    """
    Visualized point cloud.

    Output color coded point cloud. Color values indicate individual segments.

    Args:
        coord (np.ndarray): 2D or 3D array of shape [(s) x X x Y x Z] or [(s) x X x Y].
        segmented (bool): If True expect (s) in a data format as segmented values.
        rgb (np.ndarray): Optional, indicate rgb values.
        animate (bool): Optional trigger to turn off animated rotation.
    """
    coord, check = _dataset_format(coord=coord, segmented=segmented)

    if check:
        pcd = o3d.geometry.PointCloud()

        if segmented:
            pcd.points = o3d.utility.Vector3dVector(coord[:, 1:])
        else:
            pcd.points = o3d.utility.Vector3dVector(coord)

        if rgb is not None:
            if np.max(rgb) > 1:
                rgb = rgb / 255
            pcd.colors = o3d.utility.Vector3dVector(rgb)
        else:
            pcd.colors = o3d.utility.Vector3dVector(_rgb(coord, segmented))

        if animate:
            o3d.visualization.draw_geometries_with_animation_callback(
                [pcd], rotate_view
            )
        else:
            o3d.visualization.draw_geometries_with_animation_callback(
                [pcd], background_view
            )


def rotate_view(vis):
    """
    Optional viewing parameter for open3D to constantly rotate scene.
    Args:
        vis: Open3D view control setting.
    """
    ctr = vis.get_view_control()
    ctr.rotate(2.0, 0.0)

    return False


def background_view(vis):
    """
    Optional viewing parameter for open3D to constantly rotate scene.
    Args:
        vis: Open3D view control setting.
    """
    opt = vis.get_render_option()
    opt.background_color = np.asarray([0, 0, 0])
    return False


def VisualizeFilaments(coord: np.ndarray, animate=True, with_node=True):
    """
    Visualized filaments.

    Output color coded point cloud. Color values indicate individual segments.

    Args:
        coord (np.ndarray): 2D or 3D array of shape [(s) x X x Y x Z] or [(s) x X x Y].
        animate (bool): Optional trigger to turn off animated rotation.
        with_node (bool): Optional, If True show points on filament
    """
    coord, check = _dataset_format(coord=coord, segmented=True)

    if check:
        if with_node:
            pcd = o3d.geometry.PointCloud()
            pcd.points = o3d.utility.Vector3dVector(coord[:, 1:])
            pcd.colors = o3d.utility.Vector3dVector(_rgb(coord, True))

        graph = segment_to_graph(coord=coord)
        line_set = o3d.geometry.LineSet()

        line_set.points = o3d.utility.Vector3dVector(coord[:, 1:])
        line_set.lines = o3d.utility.Vector2iVector(graph)
        line_set.colors = o3d.utility.Vector3dVector(_rgb(coord, True, True))

        if animate:
            if with_node:
                o3d.visualization.draw_geometries_with_animation_callback(
                    [pcd, line_set], rotate_view
                )
            else:
                o3d.visualization.draw_geometries_with_animation_callback(
                    [line_set], rotate_view
                )
        else:
            if with_node:
                o3d.visualization.draw_geometries_with_animation_callback(
                    [pcd, line_set], background_view
                )
            else:
                o3d.visualization.draw_geometries([line_set])


def VisualizeScanNet(coord: np.ndarray, segmented: True):
    """
    Visualized scannet scene

    Output color-coded point cloud. Color values indicate individual segments.

    Args:
        coord (np.ndarray): 2D or 3D array of shape [(s) x X x Y x Z] or [(s) x X x Y].
        segmented (bool): If True expect (s) in a data format as segmented values.
    """
    coord, check = _dataset_format(coord=coord, segmented=segmented)

    if check:
        pcd = o3d.geometry.PointCloud()

        if segmented:
            pcd.points = o3d.utility.Vector3dVector(coord[:, 1:])
        else:
            pcd.points = o3d.utility.Vector3dVector(coord)
        pcd.colors = o3d.utility.Vector3dVector(_rgb(coord, segmented, True))

        o3d.visualization.draw_geometries([pcd])



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################

import inspect
import sys
from os.path import expanduser, join
from typing import Tuple, Union, Any

from tardis_em.utils.logo import TardisLogo

id_dict = {
    "0": "NO_ERROR",  # All good
    "01": "FEATURE_NOT_SUPPORTED_ASK_DEVELOPER",
    "1": "UNKNOWN_DATA_COMPATIBILITY_ERROR",  # Data compatibility
    "10": "DATA_COMPATIBILITY_BUILDING_DATASET",
    "100": "DATA_COMPATIBILITY | NO_DATA_FOUND",
    "101": "MISSING_IMAGES_OR_MASK_FILES",
    "11": "DATA_COMPATIBILITY",  # Data set compatibility and processing.
    "111": "WRONG_IMAGE_OR_MASK_AFTER_PREPROCESSING",
    "112": "TRIM_SIZE_INCOMPATIBLE_WITH_ARRAY_SIZE",
    "113": "INCORRECT_SHAPE",
    "114": "INCORRECT_DTYPE",
    "115": "EMPTY_OR_INCORRECT_ARRAY",
    "116": "INCOMPATIBLE_ARRAY_AFTER_PROCESSING",
    "117": "NOT_MATCHING_LENGTH_OF_LIST",
    "12": "EMPTY_DIRECTORY",  # General
    "121": "FILE_NOT_FOUND",
    "122": "DIRECTORY_NOT_FOUND",
    "124": "MISSING_OR_WRONG_PARAMETER",
    "130": "DATA_FORMAT_INCOMPATIBLE",  # General Utils error
    "131": "VALUE_ERROR_WHILE_LOADING_DATA",
    "132": "NUMPY_NOT_COMPATIBLE_WITH_AMIRA",
    "133": "VALUE_ERROR_WHILE_EXPORTING_DATA",
    "134": "VALUE_ERROR_WHILE_DRAWING_MASK",
    "139": "MISSING_VALUES_IN_PREDICTOR",
    "14": "FATAL_ERROR_BUILDING_CNN",  # SpindleTorch model
    "140": "INITIALIZATION_ERROR",
    "141": "UNSUPPORTED_NETWORK_NAME",
    "142": "CONVOLUTION_GROUP_NORM_ERROR",
    "143": "CONVOLUTION_PARAMETER_VALUE_ERROR",
    "145": "DATA_ERROR_ARRAY_SHAPE",
    "146": "DATA_AUGMENTATION_ERROR",
    "147": "DATA_DTYPE_ERROR",
    "151": "INCORRECT_PARAMETER_IN_MAIN",  # Main entry's
    "161": "MISSING_ARGUMENT",  # DIST model
    "19": "AWS_INCORRECT_VALUE",  # AWS
    "2": "MISSING_LIBRARY",
    "20": "PYTEST_ERROR",
}


def standard_error_id(id_: str):
    """
    Helper function to read MRC header.

    Args:
        id_ (str): Tardis error id.
    """

    try:
        return id_dict[id_]
    except KeyError:
        return "UNKNOWN_ERROR"


class TardisError(Exception):
    """
    MAIN ERROR HANDLER

    Args:
        id_ (str): Standardized error code. See more in documentation
        py (str): .py file location
        desc (str): Error description to pass to the shell
        warning_ (bool): If True return error and terminate all processes, eles
            show warning only.

    Returns:
        str: TARDIS Error log
    """

    def __init__(
        self, id_="0", py="NA", desc="Unknown exertion occurred!", warning_=True
    ):
        super().__init__()

        self.WIDTH = None
        id_desc = standard_error_id(id_)
        if id_desc == "UNKNOWN_ERROR":
            id_ = 42  # Swap error for unknown error code

        self.tardis_error_rise = TardisLogo()
        prev_frame = inspect.currentframe().f_back

        dir_ = join(expanduser("~"), "_tardis_error.log")
        with open(dir_, "w") as f:
            f.write(f"TARDIS ERROR CODE: {id_} {id_desc} \n")
            f.write(f"{prev_frame} \n")
            f.write("\n")
            f.write("Location :\n")
            f.write(f"{py} \n")
            f.write("\n")
            f.write("Description : \n")
            f.write(f"{desc} \n")

        desc_3, desc_4, desc_5, desc_6, desc_7, desc_8, desc_9, desc_10 = self.cut_desc(
            desc
        )
        code = "ERROR" if not warning_ else "WARNING"

        self.tardis_error_rise(
            title=f"TARDIS {code} CODE {id_} {id_desc} \n",
            text_1="Error accounted in:",
            text_2=f"{prev_frame.f_code.co_name}: {py}",
            text_3=desc_3,
            text_4=desc_4,
            text_5=desc_5,
            text_6=desc_6,
            text_7=desc_7,
            text_8=desc_8,
            text_9=desc_9,
            text_10=desc_10,
        )
        if not warning_:
            sys.exit()

    def cut_desc(self, desc: str) -> Tuple[str, str, str, str, str, str, str, str]:
        """
        Cut a string of text if too long to fit a shell window.

        Args:
            desc (str): Description text.

        Returns:
            list[str]: list of cut string
        """
        self.tardis_error_rise.cell_width()

        WIDTH = self.tardis_error_rise.WIDTH - 21
        if len(desc) <= WIDTH:
            text_3 = desc
            text_4, text_5, text_6, text_7, text_8, text_9, text_10 = (
                "",
                "",
                "",
                "",
                "",
                "",
                "",
            )
        else:
            (
                text_3,
                text_4,
                text_5,
                text_6,
                text_7,
                text_8,
                text_9,
                text_10,
            ) = self._truncate_str(desc, WIDTH)

        return text_3, text_4, text_5, text_6, text_7, text_8, text_9, text_10

    @staticmethod
    def _truncate_str(desc: str, width: int) -> Union[str, Any]:
        """
        Truncate string text up to 8th strings of max width fitted to the shell window.
        Args:
            desc (str): Description string.
            width (int): Shell width.

        Returns:
            Tuple[str, str, str, str, str, str, str, str]: Truncate string.
        """
        MAX_TRUNC = 8
        iter_i = 0

        text_ = []
        while MAX_TRUNC != iter_i:
            iter_i += 1

            text_.append(desc[:width])
            desc = desc[width:]

        return text_



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################
from typing import Tuple

import numpy as np
from scipy.spatial.distance import cdist

from tardis_em.analysis.filament_utils import (
    cut_150_degree,
    reorder_segments_id,
    sort_segment,
)
from tardis_em.analysis.geometry_metrics import total_length
from tardis_em.utils.errors import TardisError


class FilterConnectedNearSegments:
    """
    Connect splines based on spline trajectory and spline end distances.
    """

    def __init__(self, distance_th=1000, cylinder_radius=150):
        """
        Initialize the class with the distance threshold and cylinder radius parameters.

        Args:
            distance_th (int): Maximum distance between two spline endpoints for them
            to be considered for merging.
            cylinder_radius (int): Maximum distance between spline endpoint and the line
            connecting its neighboring endpoints for them to be considered for merging.
        """
        self.distance_th = distance_th
        self.cylinder_radius = cylinder_radius

    @staticmethod
    def _in_cylinder(point: np.ndarray, axis: tuple, r: int, h: int) -> bool:
        """
        Fast check if point is inside a cylinder by projecting point on cylinder
        volume in 3D.

        Args:
            point (np.ndarray): Point in 3D to project on cylinder.
            axis (tuple): Cylinder axis of orientation.
            r (int): Cylinder radius.
            h (int): Cylinder length.

        Returns:
            bool: If True, given point can be found in a cylinder.
        """
        # project the point onto the axis
        ax1 = (point - axis[0]) + 1e-16
        ax2 = (axis[1] - axis[0]) + 1e-16

        d = np.dot(ax1, ax2) / np.linalg.norm(ax2)

        # check if d is within the range [0, h]
        if d < 0 or d > h:
            return False

        # calculate the perpendicular distance
        d = np.linalg.norm(ax1 - d * (ax2) / np.linalg.norm(ax2))

        # check if d is less than the radius
        return d <= r

    @staticmethod
    def _remove_duplicates(d: dict) -> dict:
        """
        Remove duplicate splines

        Args:
            d (dict): Dictionary containing splines.

        Returns:
            dict: dictionary containing unique splines.
        """
        new_d = {}
        [
            new_d.update({k: v.tolist()})
            for k, v in d.items()
            if v.tolist() not in new_d.values()
        ]

        return new_d

    def splines_direction(self, spline1: list, spline2: list) -> bool:
        """
        Check if two splines facing the same direction. Knowing that spline1 facing
        a direction of spline2, check if spline2 also face a direction of spline1.

        Args:
            spline1 (list): Sorted array of point with 3D coordinates.
            spline2 (list): Sorted array of point with 3D coordinates.

        Returns:
            bool: If true, given splines facing the same direction
        """
        # Check 01 - 01 & Check 01 - 10
        ax = [
            (np.array(spline2[1]), np.array(spline2[0])),
            (np.array(spline2[1]), np.array(spline2[0])),
        ]
        points = [np.array(spline1[0]), np.array(spline1[-1])]
        s201_s101, s201_s110 = [
            self._in_cylinder(
                point=p, axis=a, r=self.cylinder_radius, h=self.distance_th
            )
            for p, a in zip(points, ax)
        ]

        # Check 10 - 01 and Check 10 - 10
        ax = [
            (np.array(spline2[-2]), np.array(spline2[-1])),
            (np.array(spline2[-2]), np.array(spline2[-1])),
        ]
        points = [np.array(spline1[0]), np.array(spline1[-1])]
        s210_s101, s210_s110 = [
            self._in_cylinder(
                point=p, axis=a, r=self.cylinder_radius, h=self.distance_th
            )
            for p, a in zip(points, ax)
        ]

        # Check if splines facing same direction in any way
        return np.any((s201_s101, s201_s110, s210_s101, s210_s110))

    def marge_splines(
        self, point_cloud: np.ndarray, omit_border: int, initial: bool
    ) -> np.ndarray:
        """
        Connect splines in the point_cloud that are close enough to each other
         and return the reordered splines.

        Example:
            - Get dictionary with {id: [sorted list of points]}

            - While there is more than 1 spline in the dict:
                - Pick first spline in a dict.
                - Select all ends
                - Calc. end distance to all other ends.

                - If any ends within threshold distance:
                    - Calculate initial end vector
                    - Calculate distance to the cylinder

                    - If any picked ends are within cylinder radius
                        - Pick one with the smallest distance
                        - Save two spline IDs to connected

                - If Connected ID list not empty:
                    - Marge and sort points from two IDs
                    - Check tortuosity

                    - If tortuosity <= 1.5:
                        - Add to the connected dict
                        - Remove from dict
                        - start over

                - If connected ID list is empty or tortuosity > 1.5:
                    - Add only initial spline to the connected dict
                    - Remove from dict
                    - start over

        Args:
            point_cloud (np.ndarray): Array with segmented and sorted point cloud
                of a shape [ID, X, Y, Z].
            omit_border (int): In A, distance from the border as a limit not to
                connect splines.
            initial (bool): Initial run for the operation.

        Returns:
            np.ndarray: Array of segmented point cloud with connected splines
            that fit the criteria
        """
        # Find Z bordered to filter out MT connection at the borders
        MIN_Z, MAX_Z = np.min(point_cloud[:, 3]), np.max(point_cloud[:, 3])

        # Create a dictionary to store spline information
        # Iterate through the point cloud and add points to their respective splines
        if initial:
            splines_list_df = {}
            for point in point_cloud:
                id_, x, y, z = point
                if id_ not in splines_list_df:
                    splines_list_df[id_] = []
                splines_list_df[id_].append([x, y, z])

            splines_list = {}
            for i in splines_list_df:
                value = splines_list_df[i]
                if len(value) > 5:
                    splines_list[i] = value
        else:
            splines_list = {}
            for point in point_cloud:
                id_, x, y, z = point
                if id_ not in splines_list:
                    splines_list[id_] = []
                splines_list[id_].append([x, y, z])

        # Iterate throw every spline in the list
        merge_splines = {}
        spline_id = 0
        while len(splines_list) > 1:
            key = list(splines_list.keys())[0]
            value = splines_list[key]  # Pick first spline in the dictionary
            end01 = value[0]
            end10 = value[-1]

            end01_list = [list(x)[1][0] for x in splines_list.items()]
            end10_list = [list(x)[1][-1] for x in splines_list.items()]

            # Check if any ends is within threshold distance
            end01_list01 = np.sqrt(
                np.sum((np.asarray(end01_list) - np.asarray(end01)) ** 2, axis=1)
            )
            end01_list01 = [
                {id_: dist}
                for id_, dist in zip(list(splines_list.keys()), end01_list01)
                if dist <= 1000 and id_ != key
            ]
            end01_list10 = np.sqrt(
                np.sum((np.asarray(end10_list) - np.asarray(end01)) ** 2, axis=1)
            )
            end01_list10 = [
                {id_: dist}
                for id_, dist in zip(list(splines_list.keys()), end01_list10)
                if dist <= self.distance_th and id_ != key
            ]
            end10_list01 = np.sqrt(
                np.sum((np.asarray(end01_list) - np.asarray(end10)) ** 2, axis=1)
            )
            end10_list01 = [
                {id_: dist}
                for id_, dist in zip(list(splines_list.keys()), end10_list01)
                if dist <= self.distance_th and id_ != key
            ]
            end10_list10 = np.sqrt(
                np.sum((np.asarray(end10_list) - np.asarray(end10)) ** 2, axis=1)
            )
            end10_list10 = [
                {id_: dist}
                for id_, dist in zip(list(splines_list.keys()), end10_list10)
                if dist <= self.distance_th and id_ != key
            ]

            # Check if any of the point is within the cylinder and get the closest one
            splines_to_merge = []
            end_lists = [end01_list01, end01_list10, end10_list01, end10_list10]

            for end_list in end_lists:
                for i in end_list:
                    m_id = list(i.keys())[
                        -1 if end_list in [end10_list01, end10_list10] else 0
                    ]
                    m_end = list(splines_list[m_id])

                    not_at_the_border = np.all(
                        [
                            (
                                m_end[
                                    (
                                        -1
                                        if end_list in [end10_list01, end10_list10]
                                        else 0
                                    )
                                ][2]
                                - MIN_Z
                            )
                            >= omit_border,
                            (
                                MAX_Z
                                - m_end[
                                    (
                                        -1
                                        if end_list in [end10_list01, end10_list10]
                                        else 0
                                    )
                                ][2]
                            )
                            >= omit_border,
                        ]
                    )

                    if not_at_the_border:
                        points = np.array(
                            m_end[-1 if end_list in [end10_list01, end10_list10] else 0]
                        )
                        axis = (
                            np.array(
                                value[
                                    (
                                        -2
                                        if end_list in [end10_list01, end10_list10]
                                        else 1
                                    )
                                ]
                            ),
                            np.array(
                                value[
                                    (
                                        -1
                                        if end_list in [end10_list01, end10_list10]
                                        else 0
                                    )
                                ]
                            ),
                        )

                        in_cylinder = self._in_cylinder(
                            point=points,
                            axis=axis,
                            r=self.cylinder_radius,
                            h=self.distance_th,
                        )
                        if in_cylinder:
                            splines_to_merge.append([m_id, m_end])

            # Check if any ends fit criteria for connection
            if len(splines_to_merge) == 1:  # Merge
                # Check if selected spline facing the same direction
                same_direction = self.splines_direction(value, splines_to_merge[0][1])

                # Connect splines
                if same_direction:
                    merged_spline = np.concatenate((value, splines_to_merge[0][1]))
                    merge_splines[spline_id] = sort_segment(merged_spline)

                    del splines_list[splines_to_merge[0][0]]
            elif len(splines_to_merge) > 1:  # If more than one find best
                if len(np.unique([x[0] for x in splines_to_merge])) == 1:
                    # Check if selected spline facing the same direction
                    same_direction = self.splines_direction(
                        value, splines_to_merge[0][1]
                    )

                    # Connect splines
                    if same_direction:
                        merged_spline = np.concatenate((value, splines_to_merge[0][1]))

                        merge_splines[spline_id] = sort_segment(merged_spline)
                        del splines_list[splines_to_merge[0][0]]
                else:
                    end_lists = {}
                    for d in np.concatenate(
                        [end01_list01, end01_list10, end10_list01, end10_list10]
                    ):
                        end_lists.update(d)

                    # Check which splines facing the same direction
                    same_direction = []
                    for d in end_lists:
                        same_direction.append(
                            self.splines_direction(value, splines_list[d])
                        )

                    # Pick splines with the smallest distance that facing same direction
                    end_lists_id = [x for x, b in zip(end_lists, same_direction) if b]
                    same_direction = np.any(same_direction)

                    if len(end_lists_id) > 0:
                        end_lists_id = min(end_lists_id, key=lambda x: end_lists[x])
                        merged_spline = np.concatenate(
                            (value, splines_list[end_lists_id])
                        )

                        merge_splines[spline_id] = sort_segment(merged_spline)
                        del splines_list[end_lists_id]
            else:  # No merge found
                same_direction = True
                merge_splines[spline_id] = sort_segment(value)

            if not same_direction:
                merge_splines[spline_id] = sort_segment(value)

            del splines_list[key]
            spline_id += 1

        # Add the last spline to the new list
        try:
            key = list(splines_list.keys())[0]
            merge_splines[spline_id] = splines_list[key]
        except IndexError:
            pass

        return np.concatenate(
            [
                np.hstack((np.repeat(id_, len(array)).reshape(-1, 1), array))
                for id_, array in merge_splines.items()
            ]
        )

    def __call__(self, point_cloud: np.ndarray, omit_border=0):
        past_l = 0
        while len(np.unique(point_cloud[:, 0])) != past_l:
            if past_l == 0:
                past_l = len(np.unique(point_cloud[:, 0]))
                point_cloud = self.marge_splines(
                    point_cloud=point_cloud, omit_border=omit_border, initial=True
                )
            else:
                past_l = len(np.unique(point_cloud[:, 0]))
                point_cloud = self.marge_splines(
                    point_cloud=point_cloud, omit_border=omit_border, initial=False
                )
        return point_cloud


class FilterSpatialGraph:
    """
    Calculate the length of each spline and distance between all splines ends.

    This class iterates over all splines in array [ID, X, Y, Z] by given ID.
    Firstly, if specified during initialization, the class calculates the distance
    between all splines ends and use it to define which splines are obviously
    broken.
    Then it calculate length of all the splines (also new one) and filter out
    ones that are too short.
    """

    def __init__(
        self,
        connect_seg_if_closer_then=1000,
        cylinder_radius=200,
        filter_short_segments=1000,
    ):
        self.connect_seg_if_closer_then = connect_seg_if_closer_then
        self.filter_short_segments = filter_short_segments

        self.marge_splines = FilterConnectedNearSegments(
            distance_th=connect_seg_if_closer_then, cylinder_radius=cylinder_radius
        )

    def __call__(self, segments: np.ndarray) -> np.ndarray:
        """
        Connect splines that have their end's in close distance and remove
        splines that are too short.

        Args:
            segments (np.ndarray): Array of points with ID label of shape [ID, X, Y, Z]

        Returns:
            np.ndarray: Filtered array of connected MTs
        """
        """Do iterative optimization split 150 degree connection / marge"""
        # Split 150 degree connections
        loop_ = True
        while loop_:
            loop_, segments = cut_150_degree(segments)

        # Connect segments with ends close to each other
        border = [np.min(segments[:, 3]), np.max(segments[:, 3])]
        border = (border[1] - border[0]) / 50

        if self.connect_seg_if_closer_then > 0:
            segments = self.marge_splines(point_cloud=segments, omit_border=border)
            segments = reorder_segments_id(segments)

        """Remove too short splines"""
        if self.filter_short_segments > 0:
            length = []
            for i in np.unique(segments[:, 0]):
                length.append(
                    total_length(segments[np.where(segments[:, 0] == int(i))[0], 1:])
                )

            length = [
                id_ for id_, i in enumerate(length) if i > self.filter_short_segments
            ]

            new_seg = []
            for i in length:
                new_seg.append(segments[np.where(segments[:, 0] == i), :])

            if len(new_seg) > 0:
                segments = np.hstack(new_seg)[0, :]
                segments = reorder_segments_id(segments)

        # Split 150 degree connections
        loop_ = True
        while loop_:
            loop_, segments = cut_150_degree(segments)

        return reorder_segments_id(segments)


class SpatialGraphCompare:
    """
    Compare two spatial graphs and output filtered-out array's of splines based
    on similarity.

    This class take as na input two arrays of shape [n, 3 or 4] for 2D or 3D
    point cloud. This arrays contain [ID x X x Y x Z] dimensions.

    The comparison is archived by calculating cdist for all splines from one spatial
    graph to all splines from second spatial graph. And for each spline it output
    probability of similarity and splines id's.

    The Probability is calculated as a ration of points (in threshold contact)
    to all points in spline.

    The selection threshold for the spline interaction is given between 0 and 1.
    """

    def __init__(self, distance_threshold: int, interaction_threshold: float):
        self.dist_th = distance_threshold
        self.inter_th = interaction_threshold

    def _compare_spatial_graphs(
        self, spatial_graph_1: np.ndarray, spatial_graph_2: np.ndarray
    ) -> list:
        """
        Wrapper to compare all MT's between two spatial graphs

        Args:
            spatial_graph_1 (np.ndarray): Spatial graph 1.
            spatial_graph_2 (np.ndarray): Spatial graph 2.

        Returns:
            list: List of MT from spatial graph 1 that match spatial graph 2.
        """
        match_sg1_sg2 = []

        for i in np.unique(spatial_graph_1[:, 0]):
            sg1_spline = spatial_graph_1[spatial_graph_1[:, 0] == i, :]
            iou = []

            for j in np.unique(spatial_graph_2[:, 0]):
                sg2_spline = spatial_graph_2[spatial_graph_2[:, 0] == j, :]
                iou.append(
                    compare_splines_probability(
                        sg1_spline[:, 1:], sg2_spline[:, 1:], self.dist_th
                    )
                )

            ids = [
                id for id, i in enumerate(iou) if np.sum(i) > 0 and i >= self.inter_th
            ]
            match_sg1_sg2.append([i, ids])

        return match_sg1_sg2

    def __call__(
        self, amira_sg: np.ndarray, tardis_sg: np.ndarray
    ) -> Tuple[list, list]:
        """
        Compute comparison of Amira and Tardis spatial graphs and output tuple of
        arrays with different selected MTs:
            - Label1: MT taken from the Tardis (matches Amira)
            - Label2: MT taken from the Amira (matches Tardis)
            - Label3: MT in Tardis without match
            - Label4: MT in Amira without match

        Args:
            amira_sg (np.ndarray): Spatial graph [ID, X, Y, Z] from Amira.
            tardis_sg (np.ndarray): Spatial graph [ID, X, Y, Z] from tardis_em.

        Returns:
            Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]: Tuple of all arrays
        """
        label = [
            "TardisFilterBasedOnAmira",
            "TardisNoise",
            "AmiraFilterBasedOnTardis",
            "AmiraNoise",
        ]

        """Amira splines scores"""
        amira_comp = self._compare_spatial_graphs(
            spatial_graph_1=amira_sg, spatial_graph_2=tardis_sg
        )

        """Compare Tardis with Amira"""
        tardis_comp = self._compare_spatial_graphs(
            spatial_graph_1=tardis_sg, spatial_graph_2=amira_sg
        )

        # Select all splines from Tardis that match Amira
        match = [x[0] for x in tardis_comp if x[1] != []]
        tardis_match = np.stack([x for x in tardis_sg if x[0] in match])

        noise = [x[0] for x in tardis_comp if x[1] == []]
        tardis_noise = np.stack([x for x in tardis_sg if x[0] in noise])

        # Select all splines from Amira that match Tardis
        match = [x[0] for x in amira_comp if x[1] != []]
        amira_match = np.stack([x for x in amira_sg if x[0] in match])

        noise = [x[0] for x in amira_comp if x[1] == []]
        amira_noise = np.stack([x for x in amira_sg if x[0] in noise])

        spatial_graphs = [
            g
            for g in (tardis_match, tardis_noise, amira_match, amira_noise)
            if g is not None
        ]
        label = [
            l
            for g, l in zip(
                (tardis_match, tardis_noise, amira_match, amira_noise), label
            )
            if g is not None
        ]

        return spatial_graphs, label


class ComputeConfidenceScore:
    @staticmethod
    def _angle_smoothness(tangents):
        """
        Calculate the smoothness of a sequence of vectors based on the angles
        between consecutive vectors.

        The smoothness is computed as 1 minus the standard deviation of the angles between
        consecutive vectors. The angles are calculated using the dot product and
        magnitude of the vectors. A higher smoothness value indicates a more
        consistent alignment between consecutive vectors, whereas a lower smoothness value
        indicates more variability in the alignment.

        Args:
            tangents (np.ndarray): A 2D numpy array of shape (n, m), where n is
            the number of vectors and m is the dimension of each vector.
            Each row in the array represents a vector.

        Returns:
            float: A scalar value representing the smoothness of the sequence of vectors.
                Ranges from 0 to 1, with higher values indicating a smoother sequence.
        """
        if (
            not isinstance(tangents, np.ndarray)
            or tangents.ndim != 2
            or tangents.shape[0] < 2
        ):
            return 1.0
        else:
            magnitudes = np.linalg.norm(tangents, axis=1)
            non_zero_vectors = tangents[magnitudes > 0]

            # All vectors are approximately zero or there's only one non-zero vector
            if non_zero_vectors.shape[0] < 2:
                return 1.0

            angles = np.arccos(
                np.einsum("ij,ij->i", tangents[:-1], tangents[1:])
                / (
                    np.linalg.norm(tangents[:-1], axis=1)
                    * np.linalg.norm(tangents[1:], axis=1)
                )
            )
            smoothness = 1 - np.std(angles)
            return smoothness

    @staticmethod
    def normalized_length(points: np.ndarray, min_l: float, max_l: float):
        """
        Calculate and normalize the total length of a sequence of points.
        The total length is calculated using the `total_length` function.
        The result is then normalized to a value between 0 and 1 based on the provided
        minimum and maximum length values.

        Args:
            points (np.ndarray):
            min_l (float): The minimum length value.
            max_l (float): The maximum length value.
        """
        length = total_length(points)
        return (length - min_l) / (max_l - min_l + 1e-16)

    def combined_smoothness(self, points: np.ndarray, min_l: float, max_l: float):
        tangents = np.diff(points, axis=0)

        scores = [
            self._angle_smoothness(tangents),
            self.normalized_length(points, min_l, max_l),
        ]

        return np.mean(scores)

    def __call__(self, segments: np.ndarray):
        if segments.shape[1] != 4:
            TardisError(
                "145",
                "tardis_em.analysis.spatial_graph_utils.py",
                f"Not segmented array. Expected shape 4 got {segments.shape[1]}",
            )
        unique_ids = np.unique(segments[:, 0])
        min_l = total_length(segments[np.where(segments[:, 0] == 0)[0], 1:])
        max_l = total_length(
            segments[np.where(segments[:, 0] == segments[-1, 0])[0], 1:]
        )

        scores = []
        for i in unique_ids:
            points = segments[np.where(segments[:, 0] == i)[0], 1:]
            scores.append(self.combined_smoothness(points, min_l, max_l))

        return scores


def compare_splines_probability(
    spline_1: np.ndarray, spline_2: np.ndarray, threshold=100
):
    """
    Compare two splines and calculate the probability of how likely given two
    splines are the same line given array of points for same or similar splines
    with no matching coordinates of points.

    Calculates the probability of two splines being similar by comparing
    the distance between their points and taking the mean of the matching
    points below a threshold.

    Parameters:
        spline_1 (np.ndarray): The first spline to compare, represented
            as an array of points.
        spline_2 (np.ndarray): The second spline to compare, represented
            as an array of points.
        threshold (int): The maximum distance between points for them to be
            considered matching.

    Returns:
        float: The probability of the splines being similar, ranging from 0.0 to 1.0.
    """
    if len(spline_1) == 0 or len(spline_2) == 0:
        return 0.0

    # Calculating distance matrix between points of 2 splines
    dist_matrix = cdist(spline_1, spline_2)

    # Check how many points on spline1 match spline2
    m_s1 = [1 if x < threshold else 0 for x in np.min(dist_matrix, axis=1)]

    # If no matching points probability is 0
    if sum(m_s1) == 0:
        return 0.0

    # Calculating intersection using % of the matching point below a threshold
    probability = sum(m_s1) / len(spline_1)

    return probability



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################

import numpy as np
import torch
import torch.nn as nn

from tardis_em.cnn.model.convolution import (
    DoubleConvolution,
    RecurrentDoubleConvolution,
)

from tardis_em.cnn.model.decoder_blocks import build_decoder
from tardis_em.cnn.model.encoder_blocks import build_encoder
from tardis_em.cnn.utils.utils import number_of_features_per_level


class BasicCNN(nn.Module):
    """
    Basic CNN MODEL

    Back-compatible with old CNN builder method. New functionality allows packaging
    the model in onnx format and rapidly re-use/deploy.

    Args:
        in_channels (int): Number of input channels for the first convolution.
        out_channels (int): Number of output channels for the last deconvolution.
        sigmoid (bool): If True, use nn.Sigmoid or nn.Softmax if False. Use True if
            nn.BCELoss is used as a loss function for (two-class segmentation).
        num_conv_layer (int): Number of convolution and deconvolution steps. A number of
            input channels for convolution is calculated as a linear progression.
            E.g. [64, 128, 256, 512].
        conv_layer_scaler (int): Scaler for the output feature channels.
        conv_kernel (int): Kernel size for the convolution.
        padding (int): Padding size for convolution.
        pool_kernel (int): kernel size for max_pooling.
        img_patch_size (int): Image patch size used for calculation network structure.
        layer_components (str): Convolution module used for building network.
        dropout (float, optional): If float, the dropout layer is built with a given drop-out rate.
        num_group (int): Number of groups for nn.GroupNorm.
        prediction (bool): If True, prediction mode is on.

    Returns if Training:
        torch.tensor: 5D torch without final activation.

    Returns if Prediction:
        If sigmoid is True:
            torch.tensor: 5D torch with final activation from nn.Sigmoid().
        If sigmoid is False:
            torch.tensor: 5D torch with final activation from nn.Softmax(dim=1).
    """

    def __init__(
        self,
        model="CNN",
        in_channels=1,
        out_channels=1,
        sigmoid=True,
        num_conv_layer=5,
        conv_layer_scaler=64,
        conv_kernel=3,
        padding=1,
        pool_kernel=2,
        img_patch_size=64,
        layer_components="3gcl",
        dropout=None,
        num_group=8,
        prediction=False,
    ):
        super(BasicCNN, self).__init__()

        self.prediction = prediction
        self.model = model
        self.encoder, self.decoder = None, None
        self.patch_sizes = img_patch_size

        self.in_channels, self.out_channels = in_channels, out_channels
        self.num_conv_layer = num_conv_layer
        self.conv_layer_scaler = conv_layer_scaler
        self.conv_kernel = conv_kernel
        self.padding = padding
        self.dropout = dropout
        self.num_group = num_group
        self.layer_components = layer_components
        self.pool_kernel = pool_kernel

        self.final_conv_layer, self.sigmoid, self.activation = None, sigmoid, None

        self.update_patch_size(img_patch_size, prediction)
        self.build_cnn_model()

    def update_patch_size(self, img_patch_size, sigmoid):
        self.prediction = sigmoid

        self.patch_sizes = [img_patch_size]
        for _ in range(self.num_conv_layer):
            img_patch_size = int(img_patch_size / 2)
            self.patch_sizes.append(img_patch_size)
        self.patch_sizes = list(reversed(self.patch_sizes))[2:]

        if not isinstance(self.model, str):
            states = self.model.state_dict()
            self.build_cnn_model()
            self.model.load_state_dict(states)

            states = None

    def build_cnn_model(self):
        """Encoder"""
        if self.model == "CNN":
            self.encoder = build_encoder(
                in_ch=self.in_channels,
                conv_layers=self.num_conv_layer,
                conv_layer_scaler=self.conv_layer_scaler,
                conv_kernel=self.conv_kernel,
                padding=self.padding,
                dropout=self.dropout,
                num_group=self.num_group,
                components=self.layer_components,
                pool_kernel=self.pool_kernel,
                conv_module=DoubleConvolution,
            )
        elif self.model == "RCNN":
            self.encoder = build_encoder(
                in_ch=self.in_channels,
                conv_layers=self.num_conv_layer,
                conv_layer_scaler=self.conv_layer_scaler,
                conv_kernel=self.conv_kernel,
                padding=self.padding,
                dropout=self.dropout,
                num_group=self.num_group,
                components=self.layer_components,
                pool_kernel=self.pool_kernel,
                conv_module=RecurrentDoubleConvolution,
            )

        """ Decoder """
        self.decoder = build_decoder(
            conv_layers=self.num_conv_layer,
            conv_layer_scaler=self.conv_layer_scaler,
            components=self.layer_components,
            conv_kernel=self.conv_kernel,
            padding=self.padding,
            sizes=self.patch_sizes,
            num_group=self.num_group,
            deconv_module=self.model,
        )

        """ Final Layer """
        if "3" in self.layer_components:
            self.final_conv_layer = nn.Conv3d(
                in_channels=self.conv_layer_scaler,
                out_channels=self.out_channels,
                kernel_size=1,
            )
        elif "2" in self.layer_components:
            self.final_conv_layer = nn.Conv2d(
                in_channels=self.conv_layer_scaler,
                out_channels=self.out_channels,
                kernel_size=1,
            )

        """ Prediction """
        if self.sigmoid:
            self.activation = nn.Sigmoid()
        else:
            self.activation = nn.Softmax(dim=1)


class UNet(BasicCNN):
    """
    2D/3D UNET MODEL

    "3D U-Net: Learning Dense Volumetric Segmentation from Sparse Annotation"
    <https://arxiv.org/pdf/1606.06650.pdf>.
    """

    def __init__(self, model="CNN", **kwargs):
        super(UNet, self).__init__(**kwargs)
        self.model = model

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward for an Unet model.

            Args:
                x (torch.Tensor): Input image features.

            Returns:
                torch.Tensor: Probability mask of predicted image.
        """
        encoder_features = []

        """ Encoder """
        for i, encoder in enumerate(self.encoder):
            x = encoder(x)

            if (len(self.encoder) - 1) != i:
                encoder_features.insert(0, x)

        """ Decoder """
        for decoder, encoder_features in zip(self.decoder, encoder_features):
            x = decoder(encoder_features, x)

        """ Prediction """
        x = self.final_conv_layer(x)

        if self.prediction:
            return self.activation(x)
        else:
            return x


class ResUNet(BasicCNN):
    """
    2D/3D RESNET MODEL

    modified of <10.1016/j.isprsjprs.2020.01.013>
    """

    def __init__(self, model="RCNN", **kwargs):
        super(ResUNet, self).__init__(**kwargs)
        self.model = model

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward for ResNet model.

            Args:
                x (torch.Tensor): Input image features.

            Returns:
                torch.Tensor: Probability mask of predicted image.
        """
        encoder_features = []

        """ Encoder """
        for i, encoder in enumerate(self.encoder):
            x = encoder(x)

            if (len(self.encoder) - 1) != i:
                encoder_features.insert(0, x)

        """ Decoder """
        for decoder, encoder_features in zip(self.decoder, encoder_features):
            x = decoder(encoder_features, x)

        """ Final layer and Prediction """
        x = self.final_conv_layer(x)

        if self.prediction:
            return self.activation(x)
        else:
            return x


class UNet3Plus(nn.Module):
    """
    3D FULLY CONNECTED UNET

    modified of <https://arxiv.org/abs/2004.08790>
    """

    def __init__(
        self,
        in_channels=1,
        out_channels=1,
        sigmoid=True,
        num_conv_layer=5,
        conv_layer_scaler=64,
        conv_kernel=3,
        padding=1,
        pool_kernel=2,
        img_patch_size=64,
        layer_components="3gcl",
        dropout=None,
        num_group=8,
        prediction=False,
        classifies=False,
        decoder_features=None,
    ):
        super(UNet3Plus, self).__init__()
        self.prediction = prediction
        self.decoder_features = decoder_features

        patch_sizes = [img_patch_size]
        for _ in range(num_conv_layer):
            img_patch_size = int(img_patch_size / 2)
            patch_sizes.append(img_patch_size)
        patch_sizes = list(reversed(patch_sizes))[2:]

        feature_map = number_of_features_per_level(conv_layer_scaler, num_conv_layer)

        """ Encoder """
        self.encoder = build_encoder(
            in_ch=in_channels,
            conv_layers=num_conv_layer,
            conv_layer_scaler=conv_layer_scaler,
            conv_kernel=conv_kernel,
            padding=padding,
            num_group=num_group,
            dropout=dropout,
            components=layer_components,
            pool_kernel=pool_kernel,
            conv_module=DoubleConvolution,
        )

        """ UNet3Plus classifier """
        if classifies:
            if "3" in layer_components:
                self.cls = nn.Sequential(
                    nn.Dropout(),
                    nn.Conv3d(
                        in_channels=feature_map[len(feature_map) - 1],
                        out_channels=2,
                        kernel_size=1,
                    ),
                    nn.AdaptiveAvgPool3d(output_size=1),
                    nn.Sigmoid(),
                )
            elif "2" in layer_components:
                self.cls = nn.Sequential(
                    nn.Dropout(),
                    nn.Conv2d(
                        in_channels=feature_map[len(feature_map) - 1],
                        out_channels=2,
                        kernel_size=1,
                    ),
                    nn.AdaptiveAvgPool2d(output_size=1),
                    nn.Sigmoid(),
                )
        else:
            self.cls = None

        """ Decoder """
        self.decoder = build_decoder(
            conv_layers=num_conv_layer,
            conv_layer_scaler=conv_layer_scaler,
            components=layer_components,
            conv_kernel=conv_kernel,
            padding=padding,
            sizes=patch_sizes,
            num_group=num_group,
            deconv_module="unet3plus",
        )

        """ Final Layer """
        if "3" in layer_components:
            self.final_conv_layer = nn.Conv3d(
                in_channels=conv_layer_scaler, out_channels=out_channels, kernel_size=1
            )
        elif "2" in layer_components:
            self.final_conv_layer = nn.Conv2d(
                in_channels=conv_layer_scaler, out_channels=out_channels, kernel_size=1
            )

        """ Prediction """
        if sigmoid:
            self.activation = nn.Sigmoid()
        else:
            self.activation = nn.Softmax(dim=1)

    @staticmethod
    def dot_product(x: torch.Tensor, x_cls: torch.Tensor) -> torch.Tensor:
        """
        Dot product for two tensors.

        Args:
            x (torch.Tensor): Image tensor.
            x_cls (torch.Tensor): Classified image tensor.

        Returns:
            torch.Tensor: Dot product of two tensors.
        """
        B, N, D, H, W = x.shape
        x = x.view(B, N, D * H * W)

        final = torch.einsum("ijk,ijk->ijk", [x, x_cls])
        final = final.view(B, N, D, H, W)

        return final

    def forward(self, x: torch.Tensor):
        """
        Forward for Unet3Plus model.

            Args:
                x (torch.Tensor): Input image features.

            Returns:
                torch.Tensor: Probability mask of predicted image.
        """
        """ Encoder """
        encoder_features = []

        for i, encoder in enumerate(self.encoder):
            x = encoder(x)

            # Don't save last encoder layer
            # Encoder layers are inverted
            if (len(self.encoder) - 1) != i:
                encoder_features.insert(0, x)

        """ Classification on last encoder """
        if self.cls is not None:
            x_cls = self.cls(x).squeeze(3).squeeze(2)
            x_cls_max = x_cls.argmax(dim=1)
            x_cls_max = x_cls_max[:, np.newaxis].float()

        """ Decoder """
        decoder_features = []
        for decoder in self.decoder:
            x = decoder(
                x=x,
                encoder_features=encoder_features,
                decoder_features=decoder_features,
            )

            # add/remove layer at each iter
            decoder_features.insert(0, x)
            encoder_features = encoder_features[1:]

        """ Final Layer / Prediction"""
        if self.cls is not None:
            x = self.dot_product(self.final_conv_layer(x), x_cls_max)

            if self.prediction:
                x = self.activation(x)

            return x, x_cls
        else:
            x = self.final_conv_layer(x)

            if self.prediction:
                return self.activation(x)
            else:
                return x


class FNet(nn.Module):
    """
    New Unet model combining Unet and Unet3Plus
    The model shares encoder path which is split for decoding patch Unet and Unet3Plus
    style. The final layers from each are summed and sigmoid

    Args:
        in_channels: Number of input channels for the first convolution.
        out_channels: Number of output channels for the last deconvolution.
        sigmoid: If True, use nn.Sigmoid or nn.Softmax if False. Use True if
            nn.BCELoss is used as a loss function for (two-class segmentation).
        num_conv_layer: Number of convolution and deconvolution steps. A number of
            input channels for convolution is calculated as a linear progression.
            E.g. [64, 128, 256, 512].
        conv_layer_scaler: Feature output of the first layer.
        conv_kernel: Kernel size for the convolution.
        padding: Padding size for convolution.
        pool_kernel: kernel size for max_pooling.
        img_patch_size: Image patch size used for calculation of network structure.
        layer_components: Convolution module used for building network.
        num_group: Number of groups for nn.GroupNorm.
        prediction: If True, prediction mode is on.
    """

    def __init__(
        self,
        in_channels=1,
        out_channels=1,
        sigmoid=True,
        num_conv_layer=5,
        conv_layer_scaler=64,
        conv_kernel=3,
        padding=1,
        pool_kernel=2,
        dropout=None,
        img_patch_size=64,
        layer_components="3gcl",
        num_group=8,
        attn_features=False,
        prediction=False,
    ):
        super(FNet, self).__init__()
        self.prediction = prediction
        self.encoder, self.decoder_unet, self.decoder_3plus = None, None, None
        self.unet_conv_layer, self.unet3plus_conv_layer, self.final_conv_layer = (
            None,
            None,
            None,
        )
        self.activation = None

        self.patch_sizes = img_patch_size

        self.in_channels, self.out_channels = in_channels, out_channels
        self.num_conv_layer = num_conv_layer
        self.conv_layer_scaler = conv_layer_scaler
        self.conv_kernel = conv_kernel
        self.padding = padding
        self.dropout = dropout
        self.num_group = num_group
        self.layer_components = layer_components
        self.pool_kernel = pool_kernel

        self.attn_features = attn_features
        self.sigmoid = sigmoid

        self.patch_sizes = [img_patch_size]
        self.update_patch_size(img_patch_size, self.prediction)
        self.build_cnn_model()

    def update_patch_size(self, img_patch_size, prediction):
        self.prediction = prediction

        self.patch_sizes = [img_patch_size]
        for _ in range(self.num_conv_layer):
            img_patch_size = int(img_patch_size / 2)
            self.patch_sizes.append(img_patch_size)
        self.patch_sizes = list(reversed(self.patch_sizes))[2:]

        if self.decoder_unet is not None:
            states = [self.decoder_unet.state_dict(), self.decoder_3plus.state_dict()]

            """ Decoder """
            self.decoder_unet = build_decoder(
                conv_layers=self.num_conv_layer,
                conv_layer_scaler=self.conv_layer_scaler,
                components=self.layer_components,
                conv_kernel=self.conv_kernel,
                padding=self.padding,
                sizes=self.patch_sizes,
                num_group=self.num_group,
            )
            self.decoder_3plus = build_decoder(
                conv_layers=self.num_conv_layer,
                conv_layer_scaler=self.conv_layer_scaler,
                components=self.layer_components,
                conv_kernel=self.conv_kernel,
                padding=self.padding,
                sizes=self.patch_sizes,
                num_group=self.num_group,
                deconv_module="unet3plus",
                attn_features=self.attn_features,
            )
            self.decoder_unet.load_state_dict(states[0])
            self.decoder_3plus.load_state_dict(states[1])

            states = None

    def build_cnn_model(self):
        """Encoder"""
        self.encoder = build_encoder(
            in_ch=self.in_channels,
            conv_layers=self.num_conv_layer,
            conv_layer_scaler=self.conv_layer_scaler,
            conv_kernel=self.conv_kernel,
            padding=self.padding,
            dropout=self.dropout,
            num_group=self.num_group,
            components=self.layer_components,
            pool_kernel=self.pool_kernel,
            conv_module=DoubleConvolution,
            attn_features=self.attn_features,
        )

        """ Decoder """
        self.decoder_unet = build_decoder(
            conv_layers=self.num_conv_layer,
            conv_layer_scaler=self.conv_layer_scaler,
            components=self.layer_components,
            conv_kernel=self.conv_kernel,
            padding=self.padding,
            sizes=self.patch_sizes,
            num_group=self.num_group,
        )
        self.decoder_3plus = build_decoder(
            conv_layers=self.num_conv_layer,
            conv_layer_scaler=self.conv_layer_scaler,
            components=self.layer_components,
            conv_kernel=self.conv_kernel,
            padding=self.padding,
            sizes=self.patch_sizes,
            num_group=self.num_group,
            deconv_module="unet3plus",
            attn_features=self.attn_features,
        )

        """ Final Layer """
        if "3" in self.layer_components:
            self.unet_conv_layer = nn.Conv3d(
                in_channels=self.conv_layer_scaler,
                out_channels=self.conv_layer_scaler,
                kernel_size=1,
            )
            self.unet3plus_conv_layer = nn.Conv3d(
                in_channels=self.conv_layer_scaler,
                out_channels=self.conv_layer_scaler,
                kernel_size=1,
            )

            self.final_conv_layer = nn.Conv3d(
                in_channels=self.conv_layer_scaler * 2,
                out_channels=self.out_channels,
                kernel_size=1,
            )
        elif "2" in self.layer_components:
            self.unet_conv_layer = nn.Conv2d(
                in_channels=self.conv_layer_scaler,
                out_channels=self.conv_layer_scaler,
                kernel_size=1,
            )
            self.unet3plus_conv_layer = nn.Conv2d(
                in_channels=self.conv_layer_scaler,
                out_channels=self.conv_layer_scaler,
                kernel_size=1,
            )

            self.final_conv_layer = nn.Conv2d(
                in_channels=self.conv_layer_scaler * 2,
                out_channels=self.out_channels,
                kernel_size=1,
            )

        """ Prediction """
        if self.sigmoid:
            self.activation = nn.Sigmoid()
        else:
            self.activation = nn.Softmax(dim=1)

    def forward(self, x: torch.Tensor):
        """
        Forward for FNet model.

            Args:
                x (torch.Tensor): Input image features.

            Returns:
                torch.Tensor: Probability mask of predicted image.
        """
        encoder_features = []

        """ Encoder """
        for i, encoder in enumerate(self.encoder):
            x = encoder(x)

            if (len(self.encoder) - 1) != i:
                encoder_features.insert(0, x)

        """ Decoders """
        x_3plus = x
        for i, (decoder, decoder_2) in enumerate(
            zip(self.decoder_unet, self.decoder_3plus)
        ):
            # Add Decoder layer
            x = decoder(encoder_features[0], x)
            x_3plus = decoder_2(
                x=x_3plus,
                encoder_features=encoder_features,
            )

            # Remove layer at each iter
            encoder_features = encoder_features[1:]

        """ Final Layer/Prediction """
        x = self.final_conv_layer(
            torch.cat(
                (self.unet_conv_layer(x), self.unet3plus_conv_layer(x_3plus)), dim=1
            )
        )

        if self.prediction:
            return self.activation(x)
        else:
            return x



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################

from typing import Optional, Tuple

import torch
from torch import nn

from tardis_em.dist_pytorch.model.modules import (
    ComparisonLayer,
    GeluFeedForward,
    PairBiasSelfAttention,
    QuadraticEdgeUpdate,
    SelfAttention2D,
    TriangularEdgeUpdate,
)


class DistStack(nn.Module):
    """
    WRAPPER FOR DIST LAYER

    This wrapper defines a number of layer for the DIST.

    Args:
        node_dim (int): Number of input dimensions for node features.
        pairs_dim (int, optional): Number of input dimensions for pairs features.
        num_layers (int): Number of GraphFormer layers. Min. 1.
        dropout (float): Dropout rate.
        ff_factor (int): Feed forward factor.
        num_heads: Number of heads in multi-head attention.
        structure (str): Define DIST structure.
    """

    def __init__(
        self,
        pairs_dim: int,
        node_dim: Optional[int] = None,
        num_layers=1,
        dropout=0,
        ff_factor=4,
        num_heads=8,
        structure="full",
    ):
        super().__init__()
        self.layers = nn.ModuleList()
        for _ in range(num_layers):
            layer = DistLayer(
                pairs_dim=pairs_dim,
                node_dim=node_dim,
                dropout=dropout,
                ff_factor=ff_factor,
                num_heads=num_heads,
                structure=structure,
            )
            self.layers.append(layer)

    def __len__(self) -> int:
        return len(self.layers)

    def __getitem__(self, i: int):
        return self.layers[i]

    def forward(
        self,
        edge_features: torch.Tensor,
        node_features: Optional[torch.Tensor] = None,
        src_mask=None,
        src_key_padding_mask=None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Forward throw individual DIST layer.

        Args:
            edge_features (torch.Tensor): Edge features as a tensor of shape
                [Batch x Length x Length x Channels].
            node_features (torch.Tensor, optional): Optional node features as a
                tensor of shape [Batch x Length x Channels].
            src_mask (torch.Tensor, optional): Optional source mask for masking
                over batches.
            src_key_padding_mask (torch.Tensor, optional): Optional mask use for
                feature padding.

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: Updated graph representation.
        """
        for layer in self.layers:
            node_features, edge_features = layer(
                h_pairs=edge_features,
                h_nodes=node_features,
                src_mask=src_mask,
                src_key_padding_mask=src_key_padding_mask,
            )

        return node_features, edge_features


class DistLayer(nn.Module):
    """
    MAIN DIST LAYER

    DistLayer takes an embedded input and performs the paired bias
    self-attention (modified multi-head attention), followed by GeLu feed-forward
    normalization to update node-embedded information. Then update from
    the GeLu is summed with the edge feature map. As an output, DIST outputs
    an attention vector for given input that encodes attention between nodes and
    pairs (edges).

    Args:
        pairs_dim (int): Output feature for pairs and nodes representation.
        node_dim (int): Input feature for pairs and nodes representation.
        dropout (float): Dropout rate.
        ff_factor (int): Feedforward factor used for GeLuFFN.
        num_heads (int): Number of heads in self-attention
        structure (str): Structure of layer ['full', 'full_af', 'self_attn',
            'triang', 'dualtriang', 'quad'].
    """

    def __init__(
        self,
        pairs_dim: int,
        node_dim: Optional[int] = None,
        dropout=0,
        ff_factor=4,
        num_heads=8,
        structure="full",
    ):
        super().__init__()
        self.pairs_dim = pairs_dim
        self.channel_dim = int(pairs_dim / 4)
        if self.channel_dim <= 0:
            self.channel_dim = 1
        self.node_dim = node_dim
        self.structure = structure
        assert self.structure in [
            "full",
            "full_af",
            "self_attn",
            "triang",
            "quad",
            "dualtriang",
        ]

        # Node optional features update
        if node_dim is not None:
            if node_dim > 0:
                self.input_attn = PairBiasSelfAttention(
                    embed_dim=node_dim, pairs_dim=pairs_dim, num_heads=num_heads
                )
                self.input_ffn = GeluFeedForward(
                    input_dim=node_dim, ff_dim=node_dim * ff_factor
                )
                self.pair_update = ComparisonLayer(
                    input_dim=node_dim, output_dim=pairs_dim, channel_dim=pairs_dim
                )

        # Edge optional MHA update
        if self.structure in ["full", "full_af", "self_attn"]:
            self.row_attention = SelfAttention2D(
                embed_dim=pairs_dim, num_heads=num_heads, dropout=dropout, axis=1
            )
            self.col_attention = SelfAttention2D(
                embed_dim=pairs_dim, num_heads=num_heads, dropout=dropout, axis=0
            )

        # Edge triangular update
        if self.structure in ["full", "full_af", "triang"]:
            self.row_update = TriangularEdgeUpdate(
                input_dim=pairs_dim, channel_dim=self.channel_dim
            )
            self.col_update = TriangularEdgeUpdate(
                input_dim=pairs_dim, channel_dim=self.channel_dim, axis=0
            )

        # Edge Optional Quadratic
        if self.structure == "quad":
            self.row_update = QuadraticEdgeUpdate(
                input_dim=pairs_dim, channel_dim=self.channel_dim
            )
            self.col_update = QuadraticEdgeUpdate(
                input_dim=pairs_dim, channel_dim=self.channel_dim, axis=0
            )

        # Edge Optional dual-triang update
        if self.structure == "dualtriang":
            self.row_update_1 = TriangularEdgeUpdate(
                input_dim=pairs_dim, channel_dim=self.channel_dim
            )
            self.col_update_1 = TriangularEdgeUpdate(
                input_dim=pairs_dim, channel_dim=self.channel_dim, axis=0
            )

            self.row_update_2 = TriangularEdgeUpdate(
                input_dim=pairs_dim, channel_dim=self.channel_dim
            )
            self.col_update_2 = TriangularEdgeUpdate(
                input_dim=pairs_dim, channel_dim=self.channel_dim, axis=0
            )

        # Edge GeLu FFN normalization layer
        self.pair_ffn = GeluFeedForward(
            input_dim=pairs_dim, ff_dim=pairs_dim * ff_factor
        )

    def update_nodes(
        self,
        h_pairs: torch.Tensor,
        h_nodes: Optional[torch.Tensor] = None,
        src_mask=None,
        src_key_padding_mask=None,
    ) -> torch.Tensor:
        """
        Transformer on the input weighted by the pair representations

        Input:
            h_paris -> Batch x Length x Length x Channels
            h_nodes -> Length x Batch x Channels

        Output:
            h_nodes -> Length x Batch x Channels

        Args:
            h_pairs (torch.Tensor): Edge features.
            h_nodes (torch.Tensor): Node features.
            src_mask (torch.Tensor): Attention mask used for mask over batch.
            src_key_padding_mask (torch.Tensor): Attention key padding mask.

        Returns:
             torch.Tensor: Updated node features.
        """
        h_nodes = h_nodes + self.input_attn(
            query=h_nodes,
            pairs=h_pairs,
            attn_mask=src_mask,
            key_padding_mask=src_key_padding_mask,
        )
        h_nodes = h_nodes + self.input_ffn(x=h_nodes)

        return h_nodes

    def update_edges(
        self,
        h_pairs: torch.Tensor,
        h_nodes: Optional[torch.Tensor] = None,
        mask: Optional[torch.Tensor] = None,
        src_key_padding_mask=None,
    ) -> torch.Tensor:
        """
        Update the edge representations based on nodes.

        Input:
            h_pairs -> Batch x Length x Length x Channels
            h_nodes -> Length x Batch x Channels

        Output:
            h_pairs -> Batch x Length x Length x Channels

        Args:
            h_pairs: Edge features.
            h_nodes: Node features.
            mask: Attention mask.
            src_key_padding_mask: Attention key padding mask.

        Returns:
            torch.Tensor: Updated edge features.
        """
        # Convert node features to edge shape
        if h_nodes is not None:
            h_pairs = h_pairs + self.pair_update(x=h_nodes)

        if src_key_padding_mask is not None:
            mask = src_key_padding_mask.unsqueeze(2) | src_key_padding_mask.unsqueeze(1)

        # Update edge features
        if self.structure == "full":
            h_pairs = (
                h_pairs
                + self.row_attention(x=h_pairs, padding_mask=mask)
                + self.col_attention(x=h_pairs, padding_mask=mask)
                + self.row_update(z=h_pairs, mask=mask)
                + self.col_update(z=h_pairs, mask=mask)
            )
        elif self.structure == "full_af":
            h_pairs = h_pairs + self.row_attention(x=h_pairs, padding_mask=mask)
            h_pairs = h_pairs + self.col_attention(x=h_pairs, padding_mask=mask)
            h_pairs = h_pairs + self.row_update(z=h_pairs, mask=mask)
            h_pairs = h_pairs + self.col_update(z=h_pairs, mask=mask)
        elif self.structure == "self_attn":
            h_pairs = (
                h_pairs
                + self.row_attention(x=h_pairs, padding_mask=mask)
                + self.col_attention(x=h_pairs, padding_mask=mask)
            )
        elif self.structure in ["triang", "quad"]:
            h_pairs = (
                h_pairs
                + self.row_update(z=h_pairs, mask=mask)
                + self.col_update(z=h_pairs, mask=mask)
            )
        elif self.structure == "dualtriang":
            h_pairs = (
                h_pairs
                + self.row_update_1(z=h_pairs, mask=mask)
                + self.col_update_1(z=h_pairs, mask=mask)
            )
            h_pairs = (
                h_pairs
                + self.row_update_2(z=h_pairs, mask=mask)
                + self.col_update_2(z=h_pairs, mask=mask)
            )

        return h_pairs + self.pair_ffn(x=h_pairs)

    def forward(
        self,
        h_pairs: torch.Tensor,
        h_nodes: Optional[torch.Tensor] = None,
        src_mask=None,
        src_key_padding_mask=None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Wrapped forward throw all DIST layers.

        Args:
            h_pairs: Pairs representation.
            h_nodes: Node feature representation.
            src_mask: Optional attention mask.
            src_key_padding_mask: Optional padding mask for attention.

        Returns:
            Tuple[torch,Tensor, torch.Tensor]:
        """
        # Update node features and convert to edge shape
        if h_nodes is not None:
            h_nodes = self.update_nodes(
                h_pairs=h_pairs,
                h_nodes=h_nodes,
                src_mask=src_mask,
                src_key_padding_mask=src_key_padding_mask,
            )

        # Update edge features
        h_pairs = self.update_edges(
            h_pairs=h_pairs, h_nodes=h_nodes, src_key_padding_mask=src_key_padding_mask
        )
        return h_nodes, h_pairs



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################

from typing import Optional, Tuple, Union

import numpy as np
from numpy import ndarray

from tardis_em.utils.errors import TardisError


class CenterCrop:
    """
    CENTER CROP ARRAY

    Rescale the image and mask to a given size for 3D [DxHxW], or 2D images [HxW].

    Args:
        size (tuple): Output size of image in DHW/HW.
    """

    def __init__(self, size: tuple):
        if len(size) not in [2, 3]:
            TardisError(
                "146",
                "tardis_em/cnn/dataset/augmentation.py",
                "Image crop supported only for 3D and 2D! " f"But {size} was given.",
            )
        self.size = size

        if len(self.size) == 2:
            self.size = (0, size[0], size[1])

    def __call__(
        self, x: np.ndarray, y: Optional[np.ndarray] = None
    ) -> Union[Tuple[np.ndarray, np.ndarray], np.ndarray]:
        """
        Call for centre crop.

        Args:
            x (np.ndarray): Image 2D or 3D arrays.
            y (np.ndarray, optional): Optional label mask 2D or 3D arrays.

        Returns:
            np.ndarray: Cropped array.
        """
        if x.ndim not in [2, 3]:
            TardisError(
                "146",
                "tardis_em/cnn/dataset/augmentation.py",
                "Image crop supported only for 3D and 2D!",
            )
        if y is not None:
            if y.ndim not in [2, 3]:
                TardisError(
                    "146",
                    "tardis_em/cnn/dataset/augmentation.py",
                    "Image crop supported only for 3D and 2D!",
                )

        if x.ndim == 3:
            d, h, w = x.shape

            down_d = int(d // 2 + self.size[0] // 2)
            up_d = int(d // 2 - self.size[0] // 2)

        else:
            h, w = x.shape

        """"Calculate padding for crop"""
        bottom_h = int(h // 2 + self.size[1] // 2)
        top_h = int(h // 2 - self.size[1] // 2)
        right_w = int(w // 2 + self.size[2] // 2)
        left_w = int(w // 2 - self.size[2] // 2)

        """Crop"""
        if y is not None:
            if x.ndim == 3 and y.ndim == 3:
                return (
                    x[up_d:down_d, top_h:bottom_h, left_w:right_w],
                    y[up_d:down_d, top_h:bottom_h, left_w:right_w],
                )
            elif x.ndim == 2 and y.ndim == 2:
                return (
                    x[top_h:bottom_h, left_w:right_w],
                    y[top_h:bottom_h, left_w:right_w],
                )
        else:
            if x.ndim == 3:
                return x[up_d:down_d, top_h:bottom_h, left_w:right_w]
            elif x.ndim == 2:
                return x[top_h:bottom_h, left_w:right_w]


class RandomFlip:
    """
    180 RANDOM FLIP ARRAY

    Perform 180 degree flip randomly in z,x or y axis for 3D or 4D

        - 0 is z axis, 1 is x axis, 2 is y axis for 3D
        - 0 is x axis, 1 is y axis for 2D
    """

    def __call__(self, x: np.ndarray, y: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Call for random flip.

        Args:
            x: 2D/3D image array.
            y: 2D/3D label mask array.

        Returns:
            np.ndarray, np.ndarray: Flipped array.
        """
        random_state = np.random.randint(0, 9)
        if x.ndim == 2 or y.ndim == 2:
            random_state = np.random.randint(0, 6)

        if random_state <= 3:
            random_state = 0
        elif 6 >= random_state > 3:
            random_state = 1
        elif 9 >= random_state > 6:
            random_state = 2

        return np.flip(x, random_state), np.flip(y, random_state)


class RandomRotation:
    """
    MULTIPLE 90 RANDOM ROTATIONS

    Perform 90, 180 or 270-degree rotations for 2D or 3D in left or right

        - 0 is 90, 1 is 180, 2 is 270
        - 0 is left, 1 is right
    """

    def __init__(self):
        """Randomize rotation"""

    def __call__(self, x: np.ndarray, y: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Call for random rotation.

        Args:
            x: 2D/3D image array.
            y: 2D/3D label mask array.

        Returns:
            np.ndarray, np.ndarray: Rotated array.
        """
        random_rotation = np.random.randint(0, 9)
        if random_rotation <= 3:
            random_rotation = 0
        elif 6 >= random_rotation > 3:
            random_rotation = 1
        elif 9 >= random_rotation > 6:
            random_rotation = 2

        random_direction = np.random.randint(0, 10)
        if random_direction <= 3:
            random_direction = 0
        else:
            random_direction = 1

        # Evaluate data structure
        if random_direction == 0:
            if x.ndim == 2:
                axis = (0, 1)
            elif x.ndim == 3:
                axis = (1, 2)
        elif random_direction == 1:
            if x.ndim == 2:
                axis = (1, 0)
            elif x.ndim == 3:
                axis = (2, 1)

        return np.rot90(x, random_rotation, axis), np.rot90(y, random_rotation, axis)


class ComposeRandomTransformation:
    """
    RANDOM TRANSFORMATION WRAPPER

    Double wrapper for image and mask to perform random transformation

    Args:
        transformations: list of transforms objects from which single
            or multiple transformations will be selected.
    """

    def __init__(self, transformations: list):
        self.transforms = transformations
        self.random_repetition = np.random.randint(1, 4)

    def __call__(self, x: np.ndarray, y: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Call for random transformation.

        Args:
            x: 2D/3D image array.
            y: 2D/3D label mask array.

        Returns:
            np.ndarray, np.ndarray: Transformed array.
        """
        # Run randomize transformation
        for _ in range(self.random_repetition):
            random_transform = np.random.randint(0, len(self.transforms))
            transform = self.transforms[random_transform]
            x, y = transform(x, y)

        return x, y


def preprocess(
    image: np.ndarray,
    transformation: bool,
    size: Optional[Union[tuple, int]] = int,
    mask: Optional[np.ndarray] = None,
    output_dim_mask=1,
) -> Union[Tuple[ndarray, ndarray], ndarray]:
    """
    Module to augment dataset.

    Args:
        image (np.ndarray): 2D/3D array with image data.
        mask (np.ndarray, optional): 2D/3D array with semantic labels.
        transformation (bool): If True, perform transformation on img and mask with
            the same random effect.
        size (tuple, int): Image size output for center crop.
        output_dim_mask (int): Number of output channel dimensions for label mask.

    Returns:
        np.ndarray: Image and optionally label mask after transformation.
    """
    # Check if image is 2D or 3D
    if image.ndim not in [2, 3]:
        TardisError(
            "146",
            "tardis_em/cnn/dataset/augmentation.py",
            "Image crop supported only for 3D and 2D!",
        )

    if isinstance(size, tuple):
        if sum(size) / len(size) == size[0]:  # Check if image has uniform size
            size = size[0]
        else:
            raise Exception

    """Resize image if image != size"""
    if image.ndim == 3:
        if image.shape != (size, size, size):
            # resize image
            crop = CenterCrop((size, size, size))
            if mask is not None:
                image, mask = crop(image, mask)
            else:
                image = crop(image)
    elif image.ndim == 2:
        if image.shape != (size, size):
            # resize image
            crop = CenterCrop((size, size))
            if mask is not None:
                image, mask = crop(image, mask)
            else:
                image = crop(image)

    """Transform image randomly"""
    if transformation:
        random_transform = ComposeRandomTransformation([RandomFlip(), RandomRotation()])
        if mask is None:
            image, _ = random_transform(image, image)
        else:
            image, mask = random_transform(image, mask)

    """Expand dimension order for HW / DHW to CHW / CDHW"""
    image = np.expand_dims(image, axis=0)

    if mask is not None:
        mask = np.expand_dims(mask, axis=0)

        if output_dim_mask != mask.shape[0]:
            mask = np.tile(mask, [output_dim_mask, 1, 1, 1])
        return image, mask
    else:
        return image



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################

from typing import Tuple, Union, List

import numpy as np
import torch

from tardis_em.dist_pytorch.datasets.augmentation import BuildGraph
from tardis_em.utils.errors import TardisError


class PatchDataSet:
    """
    BUILD PATCHED DATASET

    Main change in v0.1.0RC3
        - Build moved to 3D patches

    Class for computing optimal patch size for a maximum number of points per patch.
    The optimal size of the patch is determined by max number of points. It works by first
    calculating boundary box which is used to build 3D voxels. Voxel size is initiate
    and reduced to optimize voxel sizes fo point cloud can be cut fo patches with
    specified 'max_number_of_points'.
    In the end, patches with a smaller number of points are marge with their neighbor
    in a way that will respect 'max_number_of_points' policy.

    Output is given as a list of arrays as torch.Tensor or np.ndarray.

    Args:
        max_number_of_points (int): Maximum allowed a number of points per patch.
        overlap (float): Percentage of overlapping voxel size
        drop_rate (float): Optimizer step size for reducing the size of patches.
        graph (bool): If True output computed graph for each patch of point cloud.
        tensor (bool): If True output all datasets as torch.Tensor.
    """

    def __init__(
        self,
        max_number_of_points=500,
        overlap=0.15,
        drop_rate=None,
        graph=True,
        tensor=True,
    ):
        # Point cloud down-sampling setting
        self.DOWNSAMPLING_TH = max_number_of_points

        # Patch setting
        self.voxel = None
        self.drop_rate = drop_rate
        self.TORCH_OUTPUT = tensor
        self.GRAPH_OUTPUT = graph
        self.EXPAND = 0.1  # Expand the boundary box by 10%
        self.STRIDE = overlap  # Create 15% overlaps between patches

        # Initialization
        self.INIT_PATCH_SIZE = np.zeros((2, 3))

    def boundary_box(self, coord, offset=None) -> np.ndarray:
        """
        Utile class function to compute boundary box in 2D or 3D

        Returns:
            np.ndarray: Boundary box dimensions
        """
        # Define x,y and (z) min and m
        # ax sizes
        if coord.shape[1] == 3:
            min_x, min_y, min_z = np.min(coord, axis=0)
            max_x, max_y, max_z = np.max(coord, axis=0)
        else:
            min_x, min_y = np.min(coord, axis=0)
            max_x, max_y = np.max(coord, axis=0)
            min_z, max_z = 0, 0

        if offset is not None:
            self.EXPAND = offset

        dx = ((min_x + max_x) / 2) - min_x
        min_x, max_x = min_x - dx * self.EXPAND, max_x + dx * self.EXPAND

        dy = ((min_y + max_y) / 2) - min_y
        min_y, max_y = min_y - dy * self.EXPAND, max_y + dy * self.EXPAND

        dz = ((min_z + max_z) / 2) - min_z
        min_z, max_z = min_z - dz * self.EXPAND, max_z + dz * self.EXPAND

        dx, dy = abs(min_x - max_x), abs(min_y - max_y)
        d = dx if max(dx, dy) == 0 else dy

        self.drop_rate = d * 0.01
        return np.array([(min_x, min_y, min_z), (max_x, max_y, max_z)])

    @staticmethod
    def center_patch(bbox, voxel_size=1) -> np.ndarray:
        """
        Creates a regular grid within a bounding box.

        Args:
            bbox: list or tuple of 6 floats representing the bounding box as
                (xmin, ymin, zmin, xmax, ymax, zmax)
            voxel_size: float representing the size of each voxel

        Returns:
            np.ndarray: Of shape (N, 3) representing the center coordinates of each voxel
        """
        # Calculate the number of voxels along each axis
        n_x = int(np.ceil((bbox[1, 0] - bbox[0, 0]) / voxel_size))
        n_y = int(np.ceil((bbox[1, 1] - bbox[0, 1]) / voxel_size))
        n_z = int(np.ceil((bbox[1, 2] - bbox[0, 2]) / voxel_size)) // 2

        # Calculate the coordinates of the voxel centers
        if n_x < 2:
            n_x = 2
        x = np.linspace(bbox[0, 0] - voxel_size / 2, bbox[1, 0] + voxel_size / 2, n_x)

        if n_y < 2:
            n_y = 2
        y = np.linspace(bbox[0, 1] - voxel_size / 2, bbox[1, 1] + voxel_size / 2, n_y)

        if n_z < 2:
            n_z = 2
        z = np.linspace(bbox[0, 2] - voxel_size / 2, bbox[1, 2] + voxel_size / 2, n_z)

        xv, yv, zv = np.meshgrid(x, y, z, indexing="ij")
        voxel_centers = np.column_stack((xv.flatten(), yv.flatten(), zv.flatten()))

        return voxel_centers

    def points_in_patch(self, coord: np.ndarray, patch_center: np.ndarray) -> bool:
        """
        Utile class function for filtering point cloud and output only point in
        patch.

        Args:
            coord (np.ndarray): 3D coordinate array.
            patch_center (np.ndarray): Array (1, 3) for the given patch center.

        Returns:
            tuple(bool): Array of all points that are enclosed in the given patch.
        """
        patch_size_x = self.INIT_PATCH_SIZE[0]
        patch_size_y = self.INIT_PATCH_SIZE[1]
        patch_size_z = self.INIT_PATCH_SIZE[2]

        # Bounding box for patch center
        patch_x, patch_y, patch_z = patch_center
        patch_min_x, patch_max_x = patch_x - patch_size_x, patch_x + patch_size_x
        patch_min_y, patch_max_y = patch_y - patch_size_y, patch_y + patch_size_y
        patch_min_z, patch_max_z = patch_z - patch_size_z, patch_z + patch_size_z

        coord_idx = (
            (coord[:, 0] <= patch_max_x)
            & (coord[:, 0] >= patch_min_x)
            & (coord[:, 1] <= patch_max_y)
            & (coord[:, 1] >= patch_min_y)
            & (coord[:, 2] <= patch_max_z)
            & (coord[:, 2] >= patch_min_z)
        )

        return coord_idx

    def optimal_patches(self, coord: np.ndarray, random=False) -> List[bool]:
        """
        Main class function to compute optimal patch size.

        The function takes init stored variable and iteratively searches for voxel size
        small enough that allow for all patches to have an equal or less max number
        of points.

        Args:
            coord (np.ndarray): List of coordinates for voxelize
            random:
        """
        bbox = self.boundary_box(coord)
        if self.voxel is not None:
            voxel = self.voxel + self.drop_rate
        else:
            voxel = abs(bbox[0, 0] - bbox[1, 0]) // 10 + self.drop_rate
            if voxel == 0:
                voxel = abs(bbox[0, 0] - bbox[1, 0]) + self.drop_rate

        all_patch = []

        """ Find points index in patches """
        if random:
            if self.voxel is not None:
                voxel = self.voxel + self.drop_rate
            else:
                voxel = abs(bbox[0, 0] - bbox[1, 0]) // 10
                if voxel == 0:
                    voxel = abs(bbox[0, 0] - bbox[1, 0]) + self.drop_rate

            patch_grid = self.center_patch(bbox=bbox, voxel_size=voxel)

            x_pos = np.sort(np.unique(patch_grid[:, 0]))
            if len(x_pos) > 1:
                x_pos = (x_pos[1] - x_pos[0]) / 2
            else:
                x_pos = bbox[[1, 0]] - x_pos[0]

            y_pos = np.sort(np.unique(patch_grid[:, 1]))
            if len(y_pos) > 1:
                y_pos = (y_pos[1] - y_pos[0]) / 2
            else:
                y_pos = bbox[1, 1] - y_pos[0]

            z_pos = np.sort(np.unique(patch_grid[:, 2]))
            if len(z_pos) > 1:
                z_pos = (z_pos[1] - z_pos[0]) / 2
            else:
                z_pos = bbox[1, 2] - z_pos[0]

            self.INIT_PATCH_SIZE = [
                x_pos + (x_pos * self.STRIDE),
                y_pos + (y_pos * self.STRIDE),
                z_pos + (z_pos * self.STRIDE),
            ]

            """Find points in each patch"""
            for patch in patch_grid:
                point_idx = self.points_in_patch(coord=coord, patch_center=patch)
                all_patch.append(point_idx)
            all_patch_df = [patch for patch in all_patch if np.sum(patch) > 0]
            all_patch_bool = [
                True if np.sum(patch) > 0 else False for patch in all_patch
            ]

            """Pick random patch"""
            df_all = []
            for _ in range(10):
                # Initially random pick
                random_ = np.argwhere(all_patch_bool).flatten()[
                    np.random.choice(len(all_patch_df))
                ]
                all_patch = [
                    self.points_in_patch(coord=coord, patch_center=patch_grid[random_])
                ]

                # Check if picked voxel have more then self.mesh points
                # and less then downsample threshold
                pc_size = np.sum(all_patch[0])

                while (
                    pc_size < self.DOWNSAMPLING_TH * 0.5
                    or pc_size > self.DOWNSAMPLING_TH
                ):
                    if pc_size > self.DOWNSAMPLING_TH:
                        while pc_size > self.DOWNSAMPLING_TH:
                            self.INIT_PATCH_SIZE = [
                                self.INIT_PATCH_SIZE[0] * 0.99,
                                self.INIT_PATCH_SIZE[1] * 0.99,
                                self.INIT_PATCH_SIZE[2] * 0.99,
                            ]
                            all_patch = [
                                self.points_in_patch(
                                    coord=coord, patch_center=patch_grid[random_]
                                )
                            ]
                            pc_size = np.sum(all_patch[0])
                    elif pc_size < self.DOWNSAMPLING_TH * 0.5:
                        while pc_size < self.DOWNSAMPLING_TH * 0.8:
                            self.INIT_PATCH_SIZE = [
                                self.INIT_PATCH_SIZE[0] * 1.01,
                                self.INIT_PATCH_SIZE[1] * 1.01,
                                self.INIT_PATCH_SIZE[2] * 1.01,
                            ]
                            all_patch = [
                                self.points_in_patch(
                                    coord=coord, patch_center=patch_grid[random_]
                                )
                            ]
                            pc_size = np.sum(all_patch[0])

                    if (
                        pc_size > self.DOWNSAMPLING_TH
                        or pc_size < self.DOWNSAMPLING_TH * 0.5
                    ):
                        random_ = np.argwhere(all_patch_bool).flatten()[
                            np.random.choice(len(all_patch_df))
                        ]
                        all_patch = [
                            self.points_in_patch(
                                coord=coord, patch_center=patch_grid[random_]
                            )
                        ]
                        pc_size = np.sum(all_patch[0])
                df_all.append(all_patch[0])
            all_patch = df_all
        else:
            th = 1
            while th != 0:
                all_patch = []

                """Initialize search with new voxel size"""
                voxel = round(voxel - self.drop_rate, 1)
                patch_grid = self.center_patch(bbox=bbox, voxel_size=voxel)
                if len(patch_grid) < 2:
                    continue

                x_pos = np.sort(np.unique(patch_grid[:, 0]))
                if len(x_pos) > 1:
                    x_pos = (x_pos[1] - x_pos[0]) / 2
                else:
                    x_pos = bbox[[1, 0]] - x_pos[0]

                y_pos = np.sort(np.unique(patch_grid[:, 1]))
                if len(y_pos) > 1:
                    y_pos = (y_pos[1] - y_pos[0]) / 2
                else:
                    y_pos = bbox[1, 1] - y_pos[0]

                z_pos = np.sort(np.unique(patch_grid[:, 2]))
                if len(z_pos) > 1:
                    z_pos = (z_pos[1] - z_pos[0]) / 2
                else:
                    z_pos = bbox[1, 2] - z_pos[0]

                self.INIT_PATCH_SIZE = [
                    x_pos + (x_pos * self.STRIDE),
                    y_pos + (y_pos * self.STRIDE),
                    z_pos + (z_pos * self.STRIDE),
                ]

                """Find points in each patch"""
                for patch in patch_grid:
                    point_idx = self.points_in_patch(coord=coord, patch_center=patch)
                    all_patch.append(point_idx)

                all_patch = [patch for patch in all_patch if np.sum(patch) > 0]
                th = sum([True for p in all_patch if np.sum(p) > self.DOWNSAMPLING_TH])

            """ Combine smaller patches with threshold limit """
            new_patch = []
            while len(all_patch) > 0:
                df = all_patch[0]

                if np.sum(df) >= self.DOWNSAMPLING_TH:
                    new_patch.append(df)
                    all_patch.pop(0)
                else:
                    while np.sum(df) <= self.DOWNSAMPLING_TH:
                        if len(all_patch) == 1:
                            break
                        if np.sum(df) + np.sum(all_patch[1]) > self.DOWNSAMPLING_TH:
                            break
                        df += all_patch[1]
                        all_patch.pop(1)
                    new_patch.append(df)
                    all_patch.pop(0)
            all_patch = new_patch

        return all_patch

    @staticmethod
    def normalize_idx(coord_with_idx: np.ndarray) -> np.ndarray:
        """
        Utile class function to replace ids with ordered output ID values for
        each point in patches. In other words, it produces a standardized ID for each
        point, so it can be identified with the source.

        Args:
            coord_with_idx (np.ndarray): Coordinate id value i.
        Returns:
            np.ndarray: An array all points in a patch with corrected ID value.
        """
        unique_idx, inverse_idx = np.unique(coord_with_idx[:, 0], return_inverse=True)
        # Hot-fix for numpy 2.0.0
        if inverse_idx.ndim == 2:
            inverse_index = inverse_idx[:, 0]

        norm_idx = np.arange(len(unique_idx))

        for _, id_ in enumerate(unique_idx):
            mask = coord_with_idx[:, 0] == id_
            coord_with_idx[:, 0][mask] = norm_idx[inverse_idx[mask]]
        return coord_with_idx

    def output_format(self, data: np.ndarray) -> Union[np.ndarray, torch.Tensor]:
        """
        Utile class function to output an array in the correct format (numpy or tensor)

        Args:
            data (np.ndarray): Input data for format change.

        Returns:
            np.ndarray: Array in file format specified by self.torch_output.
        """
        if self.TORCH_OUTPUT:
            data = torch.from_numpy(data).type(torch.float32)
        return data

    def patched_dataset(
        self,
        coord: np.ndarray,
        label_cls=None,
        rgb=None,
        mesh=6,
        random=False,
        voxel_size=None,
    ) -> Union[Tuple[List, List, List, List, List], Tuple[List, List, List, List]]:
        coord_patch = []
        graph_patch = []
        output_idx = []

        if voxel_size is not None:
            self.voxel = voxel_size
        else:
            self.voxel = None

        if self.GRAPH_OUTPUT:
            if coord.shape[1] not in [3, 4]:
                TardisError(
                    "113",
                    "tardis_em/dist_pytorch/datasets/patches.py",
                    "If graph True, coord must by of shape"
                    f"[Dim x X x Y x (Z)], but is: {coord.shape}",
                )
            segmented_coord = coord
            coord = coord[:, 1:]

            if mesh > 2:
                graph_builder = BuildGraph(K=mesh, mesh=True)
            else:
                graph_builder = BuildGraph(K=mesh, mesh=False)
        else:
            graph_builder = None
            if coord.shape[1] not in [2, 3]:
                TardisError(
                    "113",
                    "tardis_em/dist_pytorch/datasets/patches.py",
                    "If graph False, coord must by of shape"
                    f"[X x Y x (Z)], but is: {coord.shape}",
                )
            segmented_coord = None
            if coord.shape[1] == 2:
                """Transform 2D coord to 3D of shape [X, Y, Z]"""
                coord = np.vstack(
                    (
                        coord[:, 0],
                        coord[:, 1],
                        np.zeros((coord.shape[0],)),
                    )
                ).T
            else:
                coord = coord

        if coord.shape[0] <= self.DOWNSAMPLING_TH:
            coord_ds = [True for _ in list(range(0, coord.shape[0], 1))]

            """ Build point cloud for each patch """
            coord_patch.append(self.output_format(coord[coord_ds, :]))

            """ Optionally - Build graph for each patch """
            if self.GRAPH_OUTPUT:
                coord_label = segmented_coord[coord_ds, :]
                coord_label = self.normalize_idx(coord_label)

                graph_patch.append(self.output_format(graph_builder(coord=coord_label)))

            """ Build output index for each patch """
            output_idx.append(np.where(coord_ds)[0])

            """ Build class label index for each patch """
            if label_cls is not None:
                cls_patch = [label_cls]
            else:
                cls_patch = [self.output_format(np.zeros((1, 1)))]

            """ Build rgb node label index for each patch """
            if rgb is not None:
                rgb_patch = [rgb]
            else:
                rgb_patch = [self.output_format(np.zeros((1, 1)))]
        else:  # Build patches for PC with max num. of point per patch
            cls_patch = []
            rgb_patch = []

            """ Find points index in patches """
            all_patch = self.optimal_patches(coord=coord, random=random)

            """Build embedded feature per patch"""
            for i in all_patch:
                """Find points and optional images for each patch"""
                df_patch_keep = i

                df_patch = coord[df_patch_keep, :]
                output_df = np.where(df_patch_keep)[0]
                coord_ds = [True for _ in list(range(0, df_patch.shape[0], 1))]

                """ Build point cloud for each patch """
                coord_patch.append(self.output_format(df_patch[coord_ds, :]))

                """ Optionally - Build graph for each patch """
                if self.GRAPH_OUTPUT:
                    segment_patch = segmented_coord[df_patch_keep, :]
                    segment_patch = self.normalize_idx(segment_patch[coord_ds, :])

                    graph_patch.append(
                        self.output_format(graph_builder(coord=segment_patch))
                    )

                """ Build output index for each patch """
                output_idx.append(output_df[coord_ds])

                """ Build class label index for each patch """
                if label_cls is not None:
                    cls_df = label_cls[df_patch_keep]
                    cls_new = np.zeros((cls_df.shape[0], 200))
                else:
                    cls_df = [0]
                    cls_new = np.zeros((1, 200))

                for id_, j in enumerate(cls_df):
                    df = np.zeros((1, 200))
                    df[0, int(j)] = 1
                    cls_new[id_, :] = df
                cls_patch.append(self.output_format(cls_new))

                """ Build rbg node label index for each patch """
                if rgb is not None:
                    rgb_df = rgb[df_patch_keep]
                else:
                    rgb_df = np.zeros((1, 1))

                rgb_patch.append(self.output_format(rgb_df))

        if self.GRAPH_OUTPUT:
            return coord_patch, rgb_patch, graph_patch, output_idx, cls_patch
        else:
            return coord_patch, rgb_patch, output_idx, cls_patch



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################

from math import sqrt
from typing import Optional

import torch
import torch.nn as nn

from tardis_em.utils.errors import TardisError


class GeLU(nn.Module):
    """
    CUSTOM GAUSSIAN ERROR LINEAR UNITS ACTIVATION FUNCTION

    Args:
        tanh (float): hyperbolic tangent value for GeLU
    """

    def __init__(self, tanh: Optional[float] = None):
        super(GeLU, self).__init__()

        if tanh is not None:
            self.tanh = sqrt(tanh)
        else:
            self.tanh = 1.4142135623730951

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward GeLu transformation.

        Args:
            x (torch.Tensor): Image tensor to transform.

        Returns:
            torch.Tensor: Transformed image tensor.
        """
        return x * 0.5 * (1.0 + torch.erf(x / self.tanh))


def convolution(
    in_ch: int,
    out_ch: int,
    components: str,
    kernel: int or tuple,
    padding: int or tuple,
    num_group=None,
) -> list:
    """
    Customizable convolution block builder.

    Build an convolution block with a specified components and order:
    dimension (2 or 3), conv (c),
    ReLu (r), LeakyReLu (l), GeLu (e), PReLu (p),
    GroupNorm (g), BatchNorm (b).

    Args:
        in_ch (int): Number of input channels.
        out_ch (int): Number of output channels.
        components (str): Components that are used for conv. block.
        kernel (int, tuple): Kernel size for the convolution.
        padding (int, tuple): Padding size for the convolution.
        num_group (int): Num. of groups for the nn.GroupNorm.
            None -> if nn.GroupNorm is not used.

    Returns:
        list: Ordered nn.Module list.
    """
    modules = []
    conv = False

    """Build CNN block"""
    for letter in components:
        """Add Convolution"""
        if "c" in letter:
            conv = True

            # Add 3DConv
            if "3" in components:
                if "g" in components or "b" in components:
                    modules.append(
                        (
                            "Conv3D",
                            nn.Conv3d(
                                in_channels=in_ch,
                                out_channels=out_ch,
                                kernel_size=kernel,
                                padding=padding,
                                bias=False,
                            ),
                        )
                    )
                else:
                    modules.append(
                        (
                            "Conv3D",
                            nn.Conv3d(
                                in_channels=in_ch,
                                out_channels=out_ch,
                                kernel_size=kernel,
                                padding=padding,
                            ),
                        )
                    )
            # Add 2DConv
            if "2" in components:
                if "g" in components or "b" in components:
                    modules.append(
                        (
                            "Conv2D",
                            nn.Conv2d(
                                in_channels=in_ch,
                                out_channels=out_ch,
                                kernel_size=kernel,
                                padding=padding,
                                bias=False,
                            ),
                        )
                    )
                else:
                    modules.append(
                        (
                            "Conv2D",
                            nn.Conv2d(
                                in_channels=in_ch,
                                out_channels=out_ch,
                                kernel_size=kernel,
                                padding=padding,
                            ),
                        )
                    )
        """Add GroupNorm"""
        if "g" == letter:
            if num_group is None:
                TardisError(
                    "142",
                    "tardis_em/cnn/model/convolution.py",
                    "Number of group is required if nn.GroupNorm is used.",
                )

            if num_group > in_ch:
                num_group = 1
            if conv:
                modules.append(
                    (
                        "GroupNorm1",
                        nn.GroupNorm(num_groups=num_group, num_channels=out_ch),
                    )
                )
            else:
                modules.append(
                    (
                        "GroupNorm2",
                        nn.GroupNorm(num_groups=num_group, num_channels=in_ch),
                    )
                )
        """Add BatchNorm"""
        if "b" == letter:
            if "3" in components:
                if conv:
                    modules.append(("BatchNorm1", nn.BatchNorm3d(out_ch)))
                else:
                    modules.append(("BatchNorm2", nn.BatchNorm3d(in_ch)))
            if "2" in components:
                if conv:
                    modules.append(("BatchNorm1", nn.BatchNorm2d(out_ch)))
                else:
                    modules.append(("BatchNorm2", nn.BatchNorm2d(in_ch)))
        """Add ReLu"""
        if "r" in letter:
            modules.append(("ReLu", nn.ReLU()))
        """Add LeakyReLu"""
        if "l" in letter:
            modules.append(("LeakyReLu", nn.LeakyReLU(negative_slope=0.1)))
        """Add GeLu"""
        if "e" in letter:
            modules.append(("GeLU", GeLU()))
        if "p" in letter:
            modules.append(("PReLU", nn.PReLU(num_parameters=out_ch)))

    return modules


class SingleConvolution(nn.Sequential):
    """
    STANDARD SINGLE 3D CONVOLUTION BLOCK

    Output single convolution composed of conv, normalization and relu in order
    defined by components variable.

    Args:
        in_ch (int): Number of input channels.
        out_ch (int): Number of output channels.
        block_type (str): Define encode or decoder path e.g.
            - 'encoder': Encoder convolution path
            - 'decoder': Decoder convolution path
        components (str): Components that are used for conv. block.
        kernel (int, tuple): Kernel size for the convolution.
        padding (int, tuple): Padding size for the convolution.
        num_group (int): Num. of groups for the nn.GroupNorm.
            None -> if nn.GroupNorm is not used.
    """

    def __init__(
        self,
        in_ch: int,
        out_ch: int,
        components: str,
        kernel: int or tuple,
        padding: int or tuple,
        num_group=None,
        block_type="any",
    ):
        super(SingleConvolution, self).__init__()

        """Build single Conv3D"""
        if "3" in components:
            conv3d = convolution(
                in_ch=in_ch,
                out_ch=out_ch,
                components=components,
                kernel=kernel,
                padding=padding,
                num_group=num_group,
            )

            for name, module in conv3d:
                self.add_module(name, module)

        """Build single Conv2D"""
        if "2" in components:
            conv2d = convolution(
                in_ch=in_ch,
                out_ch=out_ch,
                components=components,
                kernel=kernel,
                padding=padding,
                num_group=num_group,
            )

            for name, module in conv2d:
                self.add_module(name, module)


class DoubleConvolution(nn.Sequential):
    """
    DOUBLE CONVOLUTION BLOCK

    Args:
        in_ch (int): Number of input channels.
        out_ch (int): Number of output channels.
        block_type (str): Define encode or decoder path e.g.
            - 'encoder': Encoder convolution path
            - 'decoder': Decoder convolution path
        components (str): Components that are used for conv. block.
        kernel (int, tuple): Kernel size for the convolution.
        padding (int, tuple): Padding size for the convolution.
        num_group (int): Num. of groups for the nn.GroupNorm.
            None -> if nn.GroupNorm is not used.
    """

    def __init__(
        self,
        in_ch: int,
        out_ch: int,
        block_type: str,
        kernel: int or tuple,
        padding: int or tuple,
        components="cgr",
        num_group=None,
    ):
        super(DoubleConvolution, self).__init__()

        # Define in and out channels for 1st and 2nd convolutions
        if block_type not in ["encoder", "decoder"]:
            TardisError(
                "143",
                "tardis_em/cnn/model/convolution.py",
                'Only "encoder", decoder block type is supported.',
            )

        """Calculate in and out channels for double convolution"""
        if block_type == "encoder":
            conv1_in_ch, conv1_out_ch = in_ch, out_ch // 2
            conv2_in_ch, conv2_out_ch = conv1_out_ch, out_ch
        else:
            conv1_in_ch, conv1_out_ch = in_ch, out_ch
            conv2_in_ch, conv2_out_ch = out_ch, out_ch

        """Build double convolution"""
        self.add_module(
            "DoubleConv1",
            SingleConvolution(
                in_ch=conv1_in_ch,
                out_ch=conv1_out_ch,
                kernel=kernel,
                padding=padding,
                components=components,
                num_group=num_group,
            ),
        )

        self.add_module(
            "DoubleConv2",
            SingleConvolution(
                in_ch=conv2_in_ch,
                out_ch=conv2_out_ch,
                components=components,
                kernel=kernel,
                padding=padding,
                num_group=num_group,
            ),
        )


class RecurrentDoubleConvolution(nn.Module):
    """
    RECURRENT DOUBLE CONVOLUTION BLOCK

    Args:
        in_ch (int): Number of input channels.
        out_ch (int): Number of output channels.
        block_type (str): Define encode or decoder path e.g.
            - 'encoder': Encoder convolution path
            - 'decoder': Decoder convolution path
        components (str): Components that are used for conv. block.
        kernel (int, tuple): Kernel size for the convolution.
        padding (int, tuple): Padding size for the convolution.
        num_group (int): Num. of groups for the nn.GroupNorm.
            None -> if nn.GroupNorm is not used.
    """

    def __init__(
        self,
        in_ch: int,
        out_ch: int,
        block_type: str,
        kernel: int or tuple,
        padding: int or tuple,
        components="cgr",
        num_group=None,
    ):
        super(RecurrentDoubleConvolution, self).__init__()

        # Define in and out channels for 1st and 2nd convolutions
        if block_type not in ["encoder", "decoder"]:
            TardisError(
                "143",
                "tardis_em/cnn/model/convolution.py",
                'Only "encoder", decoder block type is supported.',
            )

        """Calculate in and out channels for double convolution"""
        if block_type == "encoder":
            conv1_in_ch, conv1_out_ch = in_ch, out_ch // 2
            conv2_in_ch, conv2_out_ch = conv1_out_ch, out_ch
        else:
            conv1_in_ch, conv1_out_ch = in_ch, out_ch
            conv2_in_ch, conv2_out_ch = out_ch, out_ch

        """Build RCNN block"""
        self.conv1 = SingleConvolution(
            in_ch=conv1_in_ch,
            out_ch=conv1_out_ch,
            kernel=kernel,
            padding=padding,
            components=components,
            num_group=num_group,
        )

        self.conv2 = SingleConvolution(
            in_ch=conv2_in_ch,
            out_ch=conv2_out_ch,
            components=components,
            kernel=kernel,
            padding=padding,
            num_group=num_group,
        )

        self.conv3 = SingleConvolution(
            in_ch=conv2_out_ch,
            out_ch=conv2_out_ch,
            components="c",
            kernel=kernel,
            padding=padding,
            num_group=num_group,
        )
        if "l" in components:
            self.non_linearity = nn.LeakyReLU(negative_slope=0.1, inplace=True)
        elif "r" in components:
            self.non_linearity = nn.ReLU(inplace=True)
        elif "p" in components:
            self.non_linearity = nn.PReLU()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward function for customized convolution
        Args:
            x: Image patch.

        Returns:
            torch.Tensor: Up or down convoluted image patch.
        """
        out = self.conv1(x)
        out = self.conv2(out)
        residual = out

        out = self.conv3(out)

        out += residual
        out = self.non_linearity(out)

        return out



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################

from math import ceil
from os import mkdir
from os.path import isdir, join
from typing import Optional, Tuple

import numpy as np
from tifffile import tifffile as tif

from tardis_em.cnn.data_processing.scaling import scale_image
from tardis_em.utils.device import get_device
from tardis_em.utils.export_data import to_mrc
from tardis_em.utils.errors import TardisError
from tardis_em.utils.normalization import MeanStdNormalize, RescaleNormalize


def trim_with_stride(
    image: np.ndarray,
    trim_size_xy: int,
    trim_size_z: int,
    output: str,
    image_counter: int,
    scale: list,
    clean_empty=True,
    keep_if=0.01,
    stride=25,
    mask: Optional[np.ndarray] = None,
    log=True,
    pixel_size=None,
    device=get_device("cpu"),
):
    """
    Function to patch image and mask to specified patch size with overlying area

    Output images are saved as tiff with naming shame 1_1_1_25. Where
    number indicate grid position in xyz. Last number indicate stride.

    Args:
        image (np.ndarray): Corresponding image for the labels.
        trim_size_xy (int): Size of trimming in xy dimension.
        trim_size_z (int): Size of trimming in z dimension.
        output (str): Name of the output directory for saving.
        image_counter (int): Number id of image.
        scale (tuple): Up- DownScale image and mask to the given shape or factor.
        clean_empty (bool): Remove empty patches.
        keep_if (float): If float, keep only patches that have mask.
            Evaluated based on % of pixels with mask
        stride (int): Trimming step size.
        mask (np.ndarray, None): Label mask.
        log (bool): If True, output trimming log information.
        pixel_size (None, float): If not None, save mask as mrc with pixel size information.
        device (torch.device): Optional device.
    """
    # Normalize histogram
    normalize = RescaleNormalize(clip_range=(1, 99))
    meanstd = MeanStdNormalize()

    img_dtype = np.float32

    if not isdir(join(output, "imgs")):
        mkdir(join(output, "imgs"))
    if not isdir(join(output, "masks")) and mask is not None:
        mkdir(join(output, "masks"))

    if mask is not None:
        mask_dtype = np.uint8
        image, mask, dim = scale_image(
            image=image,
            mask=mask,
            scale=tuple(scale),
            device=device,
        )
        mask = np.where(mask > 0, 1, 0).astype(np.uint8)

        if image.shape != mask.shape:
            TardisError(
                "111",
                "tardis_em/cnn/data_processing.md/trim.py",
                f"Image {image.shape} has different shape from mask {mask.shape}",
            )
    else:
        image, dim = scale_image(image=image, scale=scale, device=device)

    # Rescale image intensity
    image = normalize(meanstd(image)).astype(np.float32)
    if not image.min() >= -1 or not image.max() <= 1:  # Image not between in -1 and 1
        if image.min() >= 0 and image.max() <= 1:
            image = (image - 0.5) * 2
        elif image.min() >= 0 and image.max() <= 255:
            image = image / 255  # move to 0 - 1
            image = (image - 0.5) * 2

    if img_dtype != image.dtype:
        TardisError(
            "111",
            "tardis_em/cnn/data_processing.md/trim.py",
            f"Image {img_dtype} has different dtype after interpolation {image.dtype}",
        )

    nz, ny, nx, nc = 0, 0, 0, None
    if image.ndim == 4 and dim == 3:  # 3D RGB
        nz, ny, nx, nc = image.shape
    elif image.ndim == 3 and dim == 3:  # 2D RGB
        ny, nx, nc = image.shape
        nz = 0  # 2D
    elif image.ndim == 3 and dim == 1:  # 3D gray
        nz, ny, nx = image.shape
        nc = None  # Gray
    elif image.ndim == 2:  # 2D gray
        ny, nx = image.shape
        nc = None  # 2D
        nz = 0  # Gray

    if trim_size_xy is not None or trim_size_z is not None:
        if not nx >= trim_size_xy:
            TardisError(
                "112",
                "tardis_em/cnn/data_processing.md",
                "trim_size_xy should be equal or greater then X dimension!",
            )
        if not ny >= trim_size_xy:
            TardisError(
                "112",
                "tardis_em/cnn/data_processing.md",
                "trim_size_xy should be equal or greater then Y dimension!",
            )
    else:
        if stride is None:
            TardisError(
                "112",
                "tardis_em/cnn/data_processing.md",
                "Trim sizes or stride has to be indicated!",
            )
        trim_size_xy = 64
        trim_size_z = 64

    """Calculate number of patches and stride for xyz"""
    x, y, z = ceil(nx / trim_size_xy), ceil(ny / trim_size_xy), ceil(nz / trim_size_z)

    if nz > 0:
        x_pad, y_pad, z_pad = (
            (trim_size_xy + ((trim_size_xy - stride) * (x - 1))) - nx,
            (trim_size_xy + ((trim_size_xy - stride) * (y - 1))) - ny,
            (trim_size_z + (trim_size_z - stride) * (z - 1)) - nz,
        )
    else:
        x_pad, y_pad, z_pad = (
            (trim_size_xy + ((trim_size_xy - stride) * (x - 1))) - nx,
            (trim_size_xy + ((trim_size_xy - stride) * (y - 1))) - ny,
            0,
        )

    """Adapt number of patches or patch size for trimming"""
    if trim_size_xy is not None or trim_size_z is not None:
        while x_pad < 0:
            x += 1
            x_pad += trim_size_xy - stride
        while y_pad < 0:
            y += 1
            y_pad += trim_size_xy - stride
        if nz > 0:
            while z_pad < 0:
                z += 1
                z_pad += trim_size_z - stride
    else:
        while x_pad <= 0 or y_pad <= 0:
            trim_size_xy += 1
            x_pad = (trim_size_xy + ((trim_size_xy - stride) * (x - 1))) - nx
            y_pad = (trim_size_xy + ((trim_size_xy - stride) * (y - 1))) - ny

        if nz > 0:
            while z_pad < 0:
                trim_size_z += 1
                z_pad = (trim_size_z + ((trim_size_z - stride) * (z - 1))) - nz
        else:
            z_pad = 0

    """Expand image of a patch"""
    if nz > 0:
        if nc is not None:
            image_padded = np.pad(image, [(0, z_pad), (0, y_pad), (0, x_pad), (0, 0)])
        else:
            image_padded = np.pad(image, [(0, z_pad), (0, y_pad), (0, x_pad)])

        if mask is not None:
            mask_padded = np.pad(mask, [(0, z_pad), (0, y_pad), (0, x_pad)])
    else:
        if nc is not None:
            image_padded = np.pad(image, [(0, y_pad), (0, x_pad), (0, 0)])
        else:
            image_padded = np.pad(image, [(0, y_pad), (0, x_pad)])

            if mask is not None:
                mask_padded = np.pad(mask, [(0, y_pad), (0, x_pad)])

    """Trim image and mask with stride"""
    z_start, z_stop = 0 - (trim_size_z - stride), 0
    count = len(range(x)) * len(range(y))
    if len(range(z)) > 0:
        count = count * len(range(z))
    count_save = 0
    if z == 0:
        z = 1

    for i in range(z):
        z_start = z_start + trim_size_z - stride
        z_stop = z_start + trim_size_z
        y_start, y_stop = 0 - (trim_size_xy - stride), 0

        for j in range(y):
            y_start = y_start + trim_size_xy - stride
            y_stop = y_start + trim_size_xy
            x_start, x_stop = 0 - (trim_size_xy - stride), 0

            for k in range(x):
                x_start = x_start + trim_size_xy - stride
                x_stop = x_start + trim_size_xy

                if pixel_size is None:
                    img_name = str(f"{image_counter}_{i}_{j}_{k}_{stride}.tif")
                    mask_name = str(f"{image_counter}_{i}_{j}_{k}_{stride}_mask.tif")
                else:
                    img_name = str(f"{image_counter}_{i}_{j}_{k}_{stride}.mrc")
                    mask_name = str(f"{image_counter}_{i}_{j}_{k}_{stride}_mask.mrc")

                if nc is None:
                    if nz > 0:
                        trim_img = image_padded[
                            z_start:z_stop, y_start:y_stop, x_start:x_stop
                        ]
                        if mask is not None:
                            trim_mask = mask_padded[
                                z_start:z_stop, y_start:y_stop, x_start:x_stop
                            ]
                    else:
                        trim_img = image_padded[y_start:y_stop, x_start:x_stop]
                        if mask is not None:
                            trim_mask = mask_padded[y_start:y_stop, x_start:x_stop]
                else:
                    if nz > 0:
                        trim_img = image_padded[
                            z_start:z_stop, y_start:y_stop, x_start:x_stop, :
                        ]
                        if mask is not None:
                            trim_mask = mask_padded[
                                z_start:z_stop, y_start:y_stop, x_start:x_stop
                            ]
                    else:
                        trim_img = image_padded[y_start:y_stop, x_start:x_stop, :]
                        if mask is not None:
                            trim_mask = mask_padded[
                                z_start:z_stop, y_start:y_stop, x_start:x_stop
                            ]

                trim_img = np.array(trim_img, dtype=img_dtype)

                if clean_empty and mask is not None:
                    if np.sum(trim_mask) > keep_if:
                        count_save += 1

                        if pixel_size is None:
                            tif.imwrite(
                                join(output, "imgs", img_name),
                                trim_img,
                                shape=trim_img.shape,
                            )
                            tif.imwrite(
                                join(output, "masks", mask_name),
                                np.array(trim_mask, dtype=mask_dtype),
                                shape=trim_mask.shape,
                            )
                        else:
                            to_mrc(trim_img, pixel_size, join(output, "imgs", img_name))
                            to_mrc(
                                np.array(trim_mask, mask_dtype),
                                pixel_size,
                                join(output, "masks", mask_name),
                            )
                else:
                    count_save += 1
                    if mask is None:
                        if pixel_size is None:
                            tif.imwrite(
                                join(output, "imgs", img_name),
                                trim_img,
                                shape=trim_img.shape,
                            )
                        else:
                            to_mrc(trim_img, pixel_size, join(output, "imgs", img_name))

                    else:
                        if pixel_size is None:
                            tif.imwrite(
                                join(output, "imgs", img_name),
                                trim_img,
                                shape=trim_img.shape,
                            )
                            tif.imwrite(
                                join(output, "masks", mask_name),
                                np.array(trim_mask, dtype=mask_dtype),
                                shape=trim_mask.shape,
                            )
                        else:
                            to_mrc(trim_img, pixel_size, join(output, "imgs", img_name))
                            to_mrc(
                                np.array(trim_mask, mask_dtype),
                                pixel_size,
                                join(output, "masks", mask_name),
                            )

    if log:
        return [count, count_save]


def trim_label_mask(
    points: np.ndarray, image: np.ndarray, label_mask: np.ndarray
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    ! DEPRECIATED ! Module to trim image and mask to boundary box of point cloud.

    Args:
        points (np.ndarray): 3D coordinates of pitons.
        image (np.ndarray): corresponding image for the labels.
        label_mask (np.ndarray): empty label mask.
    """
    max_x, min_x = max(points[:, 0]), min(points[:, 0])
    max_y, min_y = max(points[:, 1]), min(points[:, 1])
    max_z, min_z = max(points[:, 2]), min(points[:, 2])

    if min_z < 0:
        min_z = 0

    image_trim = image[
        int(min_z) : int(max_z), int(min_y) : int(max_y), int(min_x) : int(max_x)
    ]

    label_mask_trim = label_mask[
        int(min_z) : int(max_z), int(min_y) : int(max_y), int(min_x) : int(max_x)
    ]

    points[:, 0] = points[:, 0] - min_x
    points[:, 1] = points[:, 1] - min_y
    points[:, 2] = points[:, 2] - min_z

    return image_trim, label_mask_trim, points
